#include <string.h> 
#include <mex.h>
#include <matrix.h>
#include <math.h>


int debugEnable = 0;

#define debugPrint if (debugEnable != 0) mexPrintf


#define PIXEL_PTR    (pImg, i, j, nStep, type) ((type*)((mxUINT8_CLASS*)pImg+(i)*nStep)+(j))
#define PIXEL_PTR_8U (pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxUINT8_CLASS)
#define PIXEL_PTR_16U(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxUINT16_CLASS)
#define PIXEL_PTR_32U(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxUINT32_CLASS)
#define PIXEL_PTR_64U(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxUINT64_CLASS)
#define PIXEL_PTR_32F(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxSINGLE_CLASS)
#define PIXEL_PTR_64F(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxDOUBLE_CLASS)

/* This function computes the RMS error between two images
Source reference image type must be uint8 and must have three channel
source row and column maps images must by same size as source refernce image
source target image can be of unique size
Output image type is single precision and single channel, and is the same size
  as the source refernce image
*/

/*******************************************************************************/
/* mexFUNCTION                                                                 */
/* Gateway routine for use with MATLAB.                                        */
/*******************************************************************************/
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
          float          *pOutputCSH_Ann32f;
          unsigned short *pInputRowMap16u,  *pInputColMap16u;
          unsigned short *pTopLeftOrg, *pBotRightMarg, *pCSH_Ann;
          unsigned char  *pVideoA, *pVideoB;
          
          unsigned short nPropRow16u, nPropCol16u;  // Indices of propagation
          
          int nErrorAccum, nElementSqrDiff, n_dims_A, n_dims_pCSH_Ann, *dims_imA, *dims_imB, *dims_pCSH_Ann;
          
          mxArray *pMxArray[1];
          
          int i, j, k, colorInd, currColA, currColB, currRowA, currRowB, currFrameA, currFrameB;
		  int currFlatIndexA, currFlatIndexB, pixelErrCurrFlatIndexA, pixelErrCurrFlatIndexB;
          int m, n, f, mTarget, nTarget, fTarget, CSH_AnnJumpIndex;
          int dimsErrOutputMat[3], *pKernelSize, KernelSize;
		  
          if (nrhs < 3 || nrhs > 4 || nlhs!=1)
                    mexErrMsgTxt("Usage: outputImage=GetANNError_C3_VIDEO(videoA, videoB, CSH_Ann, kernel_size(opt));\n");
          
		  if (mxGetClassID(prhs[0]) != mxUINT8_CLASS || mxGetClassID(prhs[1]) != mxUINT8_CLASS || mxGetClassID(prhs[2]) != mxUINT16_CLASS)
			mexErrMsgTxt("input and target images must be uint8 matrices, row and column mapping images must be a uint16 matrices, top-left and bottom-right of ROI must be a uint16 scalars and kernel size must be a uint16 scalar!");

          // Get inputs
          pVideoA    = (unsigned char*)mxGetPr(prhs[0]);
          pVideoB = (unsigned char*)mxGetPr(prhs[1]);
          pCSH_Ann = (unsigned short*)mxGetPr(prhs[2]);
		  if (nrhs ==3)
			KernelSize = 4;
		  else{
			pKernelSize = (int *)mxGetPr(prhs[3]);
			KernelSize = pKernelSize[0];
		  }
		  if (KernelSize != 4) // 4 is the only kernel_size supported for now
			mexErrMsgTxt("The only Kernel size supported for now is 4!");
		  		  
		  
		  // Check dims of Video A and CSH_Ann mat
		  n_dims_A = mxGetNumberOfDimensions(prhs[0]);
		  n_dims_pCSH_Ann = mxGetNumberOfDimensions(prhs[2]);
		   if (n_dims_A != 4 || n_dims_pCSH_Ann != 4)
		  {
			  mexErrMsgTxt("Video must be 3 dimentions, and ASH_ann must be 4 dimentions!");
		  }
      
          dims_imA = (int*)mxGetDimensions(prhs[0]);
		  m = dims_imA[0]; // Width
          n = dims_imA[1]; // Height
		  f = dims_imA[2]; // num of frames
		  if (dims_imA[3] != 3)
		  {
			  mexErrMsgTxt("Video A must be a color video, i.e. 3 color channels!");
		  }
		  dims_pCSH_Ann = (int*)mxGetDimensions(prhs[2]);
		  if (dims_pCSH_Ann[0] != m || dims_pCSH_Ann[1] != n || dims_pCSH_Ann[2] != f || dims_pCSH_Ann[3] != 3)
		  {
			  mexErrMsgTxt("CSH_Ann has wrong dimensions, it must be [m, n, f, 3] matrix, when m, n and f are dimensions of VideoA!");
		  }
		      
		  dims_imB = (int*)mxGetDimensions(prhs[1]);
          mTarget = dims_imB[0]; // Width target
          nTarget = dims_imB[1];; // Height target
		  fTarget = dims_imB[2];; // Frames Number target
          if (dims_imB[3] != 3)
		  {
			  mexErrMsgTxt("Video B must be a color video, i.e. 3 color channels!");
		  }
                  
          // Create output matrix
		  dimsErrOutputMat[0] = m-KernelSize+1; dimsErrOutputMat[1] = n-KernelSize+1; dimsErrOutputMat[2] = f-KernelSize+1;
		  plhs[0] = mxCreateNumericArray(3, dimsErrOutputMat, mxSINGLE_CLASS, false);
		  pOutputCSH_Ann32f = (float*)mxGetPr(plhs[0]);
          
		  CSH_AnnJumpIndex = m*n*f;
          
		  for (currColA=0; currColA < m-KernelSize+1; currColA++) {         
			for (currRowA=0; currRowA < n-KernelSize+1; currRowA++) {
				for (currFrameA=0; currFrameA < f-KernelSize+1; currFrameA++) {
                    currFlatIndexA = currColA + currRowA*m + currFrameA*m*n;
					currColB =		pCSH_Ann[currFlatIndexA + 0*CSH_AnnJumpIndex] -1;          
					currRowB =		pCSH_Ann[currFlatIndexA + 1*CSH_AnnJumpIndex] -1;
					currFrameB =	pCSH_Ann[currFlatIndexA + 2*CSH_AnnJumpIndex] -1;
					currFlatIndexB = currColB + currRowB*mTarget + currFrameB*mTarget*nTarget;
					
	                nErrorAccum = 0;
					for (i = 0; i < KernelSize; i++) {		
                        for (j = 0; j < KernelSize; j++) {
							for (k = 0; k < KernelSize; k++) {
								for (colorInd = 0; colorInd<3; colorInd++) {
									pixelErrCurrFlatIndexA = currFlatIndexA + i + j*m + k*m*n + colorInd*m*n*f;
									pixelErrCurrFlatIndexB = currFlatIndexB + i + j*mTarget + k*mTarget*nTarget + colorInd*mTarget*nTarget*fTarget;
								
									nElementSqrDiff = (int)pVideoA[pixelErrCurrFlatIndexA] - (int)pVideoB[pixelErrCurrFlatIndexB];
									nErrorAccum += nElementSqrDiff*nElementSqrDiff;
								}
							}
						}
					}
					pOutputCSH_Ann32f[currColA + currRowA*dimsErrOutputMat[0] + currFrameA*dimsErrOutputMat[0]*dimsErrOutputMat[1]] = (float)sqrt((float)nErrorAccum );// / (float)nNumOfROIElements;
				}
			}
		  }
}

								
								