#include <string.h> 
#include <mex.h>
#include <matrix.h>

int debugEnable = 0;

/*#define debugPrint if (debugEnable != 0) mexPrintf


#define PIXEL_PTR    (pImg, i, j, nStep, type) ((type*)((mxUINT8_CLASS*)pImg+(i)*nStep)+(j))
#define PIXEL_PTR_8U (pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxUINT8_CLASS)
#define PIXEL_PTR_16U(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxUINT16_CLASS)
#define PIXEL_PTR_32U(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxUINT32_CLASS)
#define PIXEL_PTR_64U(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxUINT64_CLASS)
#define PIXEL_PTR_32F(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxSINGLE_CLASS)
#define PIXEL_PTR_64F(pImg, i, j, nStep) PIXEL_PTR(pImg, i, j, nStep, mxDOUBLE_CLASS)
*/
/* This function computes a DC rectangle kernel of source size for the source image 
The function does not assume zero-padded edges exist, but for reasonable results these should be considered
The result image is the rectangle DC kernel of the source image, anchored on the top-left pixel of the kernel
The size of the resulting matrix is the same as the source image
Source image type must be uint8 and must be single channel
Output image type is int16
*/

/*******************************************************************************/
/* mexFUNCTION                                                                 */
/* Gateway routine for use with MATLAB.                                        */
/*******************************************************************************/
void
mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
    int        i, j, k, n_dims, n_dims_kernel, ii, jj, kk;
	int	*dims, *dims_kernel;
    unsigned char	*pInputVideo8u;
	char	*Kernel_mat;
	short	*pOutputVideo16s;
	int KernelSize = 4;
    /* Examine input (right-hand-side) arguments. */
    /*mexPrintf("\nThere are %d right-hand-side argument(s).", nrhs);
    for (i=0; i<nrhs; i++)  
	{
		mexPrintf("\n\tInput Arg %i is of type:\t%s ",i,mxGetClassName(prhs[i]));
    }
    /* Examine output (left-hand-side) arguments. 
    mexPrintf("\n\nThere are %d left-hand-side argument(s).\n", nlhs);
    */
	if (nrhs != 2 || nlhs!=1 || mxGetClassID(prhs[0]) != mxUINT8_CLASS || mxGetClassID(prhs[1]) != mxINT8_CLASS)
    {
		mexErrMsgTxt("Usage: you must gine me a UINT8 video mat, and an INT8 kernel mat;\n");
	}
	// Get inputs
	pInputVideo8u = (unsigned char*)mxGetPr(prhs[0]);
	Kernel_mat   = (char*)mxGetPr(prhs[1]);

    // Get sizes of variables (Opposite on purpose)//
	n_dims = mxGetNumberOfDimensions(prhs[0]);
	n_dims_kernel = mxGetNumberOfDimensions(prhs[1]);
	if (n_dims != 3 || n_dims_kernel != 3)
	{
	  mexErrMsgTxt("Input must be video and kernel with 3 dimentions!");
	}
	dims = (int*)mxGetDimensions(prhs[0]);
	dims_kernel = (int*)mxGetDimensions(prhs[1]);
	//mexPrintf("dims_kernel = %d, %d, %d\n", dims_kernel[0], dims_kernel[1], dims_kernel[2]);
	if (dims_kernel[0] != KernelSize || dims_kernel[1] != KernelSize || dims_kernel[2] != KernelSize)
	{
	  mexErrMsgTxt("Kernel must be %dx%dx%d matrix!",KernelSize,KernelSize,KernelSize);
	}

	// Create output matrix
	plhs[0] = mxCreateNumericArray(n_dims, dims, mxINT16_CLASS, false);
	pOutputVideo16s = (short*)mxGetPr(plhs[0]);

	for (k=0; k<dims[2]-KernelSize+1; k++)  
		for (j=0; j<dims[1]-KernelSize+1; j++)
			for (i=0; i<dims[0]-KernelSize+1; i++)
			{
				for (ii=0; ii<KernelSize; ii++)
					for (jj=0; jj<KernelSize; jj++)
						for (kk=0; kk<KernelSize; kk++)
							pOutputVideo16s[i+j*dims[0]+k*dims[0]*dims[1]] += 
									(pInputVideo8u[(i+ii)+(j+jj)*dims[0]+(k+kk)*dims[0]*dims[1]]	*	Kernel_mat[ii + jj*4 + kk*4*4]);
				//mexPrintf("\tElement %d, %d, %d is index %d and value %d .\n", 1+i, 1+j, 1+k, i+j*dims[0]+k*dims[0]*dims[1], pInputVideo8u[i+j*dims[0]+k*dims[0]*dims[1]]);
				//pOutputVideo16s[i+j*dims[0]+k*dims[0]*dims[1]] = pInputVideo8u[i+j*dims[0]+k*dims[0]*dims[1]];
			}		
}
