% Plot error image of mapping
function [Regular_errImg, Regular_CSH_MeansRMS,Normalized_errImg, Normalized_CSH_MeansRMS] = ...
        ComparePatchBasedResults(A,B,width,K_of_KNN,bMask,patch_mode,patch_params,interactive,...
                            Regular_CSH_Mapping,Regular_A_patch,Regular_B_patch,...
                            Normalized_CSH_Mapping,Normalized_A_patch,Normalized_B_patch)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PlotExampleResults - Plot results from example run
% A and B are the input images
% Regular_CSH_Mapping is the nn mapping of the CSH run
% Width is the used patch width

descriptor_mode = 0;

if (exist('patch_mode','var') && patch_mode)

    hA = patch_params.hA;
    wA = patch_params.wA;
    dA = patch_params.dA;
    hB = patch_params.hB;
    wB = patch_params.wB;
    dB = patch_params.dB;

    % regular - calc errors
    Regular_CSH_MeansRMS = 0;
    Regular_errImg = zeros(hA,wA);
    
    for t = 1:K_of_KNN
        [errImg_tmp,CSH_MeansRMS_tmp] = ...
            GetErrorMeanAndImage(Regular_CSH_Mapping(:,:,:,t),hB,wB,hA,wA,width,A,B,width,descriptor_mode,...
                                            patch_mode,Regular_A_patch,Regular_B_patch,patch_params);% Mapping error calculation
        Regular_CSH_MeansRMS = Regular_CSH_MeansRMS + CSH_MeansRMS_tmp;
        Regular_errImg = Regular_errImg + errImg_tmp;
    end
    Regular_errImg = Regular_errImg./K_of_KNN;
    Regular_CSH_MeansRMS = Regular_CSH_MeansRMS / K_of_KNN;
    
    % normalized - calc errors
    Normalized_CSH_MeansRMS = 0;
    Normalized_errImg = zeros(hA,wA);
    
    for t = 1:K_of_KNN
        [errImg_tmp,CSH_MeansRMS_tmp] = ...
            GetErrorMeanAndImage(Normalized_CSH_Mapping(:,:,:,t),hB,wB,hA,wA,width,A,B,width,descriptor_mode,...
                                            patch_mode,Normalized_A_patch,Normalized_B_patch,patch_params);% Mapping error calculation
        Normalized_CSH_MeansRMS = Normalized_CSH_MeansRMS + CSH_MeansRMS_tmp;
        Normalized_errImg = Normalized_errImg + errImg_tmp;
    end
    Normalized_errImg = Normalized_errImg./K_of_KNN;
    Normalized_CSH_MeansRMS = Normalized_CSH_MeansRMS / K_of_KNN;
    
    % now the GUI
    if (exist('interactive') && interactive)
        Regular_dA_patch = double(Regular_A_patch);
        Regular_dB_patch = double(Regular_B_patch);
        Normalized_dA_patch = double(Normalized_A_patch);
        Normalized_dB_patch = double(Normalized_B_patch);

        figure;
        subplot(4,3,[1,4]); imshow(A); title('image A');
        subplot 432; imagesc(Regular_errImg); axis off; hold on; 
            title(['Regular avg error: ' num2str(Regular_CSH_MeansRMS,'%.0f')]); axis equal; axis tight
        subplot 433; imshow(B); title('image B');
        subplot 435; imagesc(Normalized_errImg); axis off; hold on; 
            title(['Normalized avg error: ' num2str(Normalized_CSH_MeansRMS,'%.0f')]); axis equal; axis tight
        subplot 436; imshow(B); title('image B');
        
        colors = {'r','g','b','c','y','m'};
        colors = {'g'};
        while 1
            subplot(4,3,[1,4]);
            [x,y,b] = ginput(1);
            if (b==3)
                break
            end
            x = floor(x); y = floor(y);
            rColor = colors{randi(length(colors),1)};
            hold off; imshow(A); title('image A'); hold on; % clean old stuff
            plot(x,y,['*' rColor]);
            
            % Regular - draw match in B
            xB = Regular_CSH_Mapping(y,x,1);
            yB = Regular_CSH_Mapping(y,x,2);
            subplot 433;
            hold off; imshow(B); title('image B'); hold on; % clean old stuff
            plot(xB,yB,['*' rColor]);
            
            % and the patches themselvels
            indA = (y-1)*wA + x;
            indB = (yB-1)*wB + xB;
            
            patchA = reshape(Regular_dA_patch(:,indA),[width,width,dA]);
            patchB = reshape(Regular_dB_patch(:,indB),[width,width,dB]);
            
            % don't need this for regular
%             for dim = 1 : dA
%                 patchA(:,:,dim) = patchA(:,:,dim)';
%                 patchB(:,:,dim) = patchB(:,:,dim)';
%             end
            
            absdiffPatch = abs(patchA - patchB);
            ssd = norm(patchA(:)-patchB(:));

            subplot 437; imshow(uint8(patchA)); title('regular patch A');
            subplot 438; imshow(uint8(absdiffPatch)); title(['diff patch. Error is: ' num2str(ssd,'%.0f')]);
            subplot 439; imshow(uint8(patchB)); title('regular patch B');

            % Normalized - draw match in B
            xB = Normalized_CSH_Mapping(y,x,1);
            yB = Normalized_CSH_Mapping(y,x,2);
            subplot 436;
            hold off; imshow(B); title('image B'); hold on; % clean old stuff
            plot(xB,yB,['*' rColor]);
            
            % and the patches themselvels
            indA = (y-1)*wA + x;
            indB = (yB-1)*wB + xB;
            
            patchA = reshape(Normalized_dA_patch(:,indA),[width,width,dA]);
            patchB = reshape(Normalized_dB_patch(:,indB),[width,width,dB]);
            % transpose the patch
            for dim = 1 : dA
                patchA(:,:,dim) = patchA(:,:,dim)';
                patchB(:,:,dim) = patchB(:,:,dim)';
            end
            
            absdiffPatch = abs(patchA - patchB);
            ssd = norm(patchA(:)-patchB(:));

            subplot(4,3,10); imshow(uint8(patchA)); title('normalized patch A');
            subplot(4,3,11); imshow(uint8(absdiffPatch)); title(['diff patch. Error is: ' num2str(ssd,'%.0f')]);
            subplot(4,3,12); imshow(uint8(patchB)); title('normalized patch B');
        end
    end
    
    return
end

[hA,wA,dA] = size(A);
[hB,wB,dB] = size(B);

Regular_CSH_MeansRMS = 0;
Regular_errImg = zeros(hA,wA);

for t = 1:K_of_KNN
    [errImg_tmp,CSH_MeansRMS_tmp] = ...
        GetErrorMeanAndImage(Regular_CSH_Mapping(:,:,:,t),hB,wB,hA,wA,width,A,B,width,descriptor_mode);% Mapping error calculation
    Regular_CSH_MeansRMS = Regular_CSH_MeansRMS + CSH_MeansRMS_tmp;
    Regular_errImg = Regular_errImg + errImg_tmp;
end

Regular_errImg = Regular_errImg./K_of_KNN;
Regular_CSH_MeansRMS = Regular_CSH_MeansRMS / K_of_KNN;

fprintf('    CSH_nn mapping RMS error: %.3f\r\n',Regular_CSH_MeansRMS);

if (exist('patch_mode','var') && patch_mode)
    return
end

figure;
br_boundary_to_ignore = width;
Hash_error_img = Regular_errImg(1:end - br_boundary_to_ignore,1:end - br_boundary_to_ignore);
pct98 = prctile(Hash_error_img(:),99);
Hash_error_img = min(pct98,Regular_errImg(1:end - br_boundary_to_ignore,1:end - br_boundary_to_ignore));
[I,J] = bounds(Hash_error_img(:),1);
[clims(1),clims(2)] = bounds([I,J],1);
clims(2) = max(clims(1) + 1, clims(2));
subplot(221);imshow(A); title('image A');
subplot(222);imshow(B); title('image B');
subplot(223);imagesc(Hash_error_img,clims); title(['CSH error: ' num2str(Regular_CSH_MeansRMS)]);

if (exist('bMask','var') && ~isEmpty(bMask))
    bMask(bMask>0)=512; % For representation only
    subplot(224);imagesc(bMask,clims); title('Binary mask (red = hole)');
end






