function [bitCountPerTable,...        % denotes how many bits were used to build this code
    numKernelsPerAlternative,...    % this is for statistics (internal)
    binMapsForUse,filters] = ...
    BuildWalshHadamardCodes_VIDEO1(tableInd,width,currFiltImgs_A, currFiltImgs_B, ...
    visualizeFilterDistributions,numTables,hA,wA,hB,wB,insideInfo,ColorChannels,WHK_with_Cb_Cr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% A] INITIALIZATIONS
maxFilterRange = 2*256*width*width*width+1;%always odd
minFilterValue = -floor(maxFilterRange/2);

% Filters defined by:
% filters_ind = [  ...x   y   t   color % index 
%                     0   0   0   1   % 1   DC
%                     0   0   0   2   % 2
%                     0   0   0   3   % 3
%                     0   0   1   1   % 4   (0,0,1)
%                     0   0   1   2   % 5
%                     0   0   1   3   % 6
%                     0   1   0   1   % 7   (0,1,0)
%                     0   1   0   2   % 8
%                     0   1   0   3   % 9
%                     1   0   0   1   % 10  (1,0,0)
%                     1   0   0   2   % 11
%                     1   0   0   3   % 12
%                     1   1   0   1   % 13  (1,1,0) only Gray
%                     1   0   1   1   % 14  (1,0,1) only Gray
%                     0   1   1   1   % 15  (0,1,1) only Gray
%                     1   1   1   1   % 16  (1,1,1) only Gray
%                     0   0   2   1   % 17  (0,0,2) only Gray
%                     0   2   0   1   % 18  (0,2,0) only Gray
%                     2   0   0   1   % 19  (2,0,0) only Gray
%                     2   1   0   1   % 20-25   6 permutations of (1,2,0) only Gray
%                     2   0   1   1   % ...
%                     1   2   0   1
%                     1   0   2   1
%                     0   1   2   1
%                     0   2   1   1
%                     ];
    
% 'hand-crafting' the number of bins per filter
% This is done according to distance of the WH filter from the origin (empirical)

% For VIDEO: the only supported width is 4
switch width
    case 4
        % these are the WH indices of the 7 kernels we want to work with
        filters =           [1, 2,  3,  4,  7,  10, 13, 14, 15  ];
        bits_for_filters =  [5, 2,  2,  3,  3,  3,  1,  1,  1   ];
    otherwise
        error(['Patch width ' num2str(width) 'is not supported. Only 4, for now please...']);
end

if (ColorChannels == 1) % Gray level images, delete Cb/Cr kernels
    error('Gray level image is not handdles for VIDEO yet');
end


num_of_filters_for_coding = length(filters);

binMapsForUse = zeros(num_of_filters_for_coding,maxFilterRange,'uint32'); % c style ranging from 0 to numBins-1 (arbitrary initialization)

nextBit = 1;

debugDisp(insideInfo,['>>>>>>>>> Coding table number ' num2str(tableInd)   ' >>>>>>>']);


%% A] BUILDING OF CODES for A and B (iterating through the kernels)
% at each iteration we append to the code the bin index
% of the patch projected on the current WH code

for filter_ind = 1:num_of_filters_for_coding
    current_filter = filters(filter_ind);
    % retrieve current filter applied on A / B (taken from all frames of the videos A and B)
    currFiltImg_A = [];
    numFramesA = length(currFiltImgs_A);
    for frame_ind = 1 : numFramesA
        CFI = currFiltImgs_A{frame_ind}{current_filter};
        currFiltImg_A = [currFiltImg_A CFI]; %#ok<AGROW>
    end
        
    currFiltImg_B = [];
    numFramesB = length(currFiltImgs_B);
    for frame_ind = 1 : numFramesB
        CFI = currFiltImgs_B{frame_ind}{current_filter};
        currFiltImg_B = [currFiltImg_B CFI]; %#ok<AGROW>
    end
        
    numBins = 2^bits_for_filters(filter_ind);
    
    % 1) Build bin map with random offset - this is a lookup-table for binning (bin edges).
    %    --------------------------------
    %           This is done by taking a relatively small joint sample of patches from A and B,
    %           and placing bin edges such that the sample distributes evenly into the bins. In
    %           addition, bins are shifted at a random amount (first and last bin won't contain the same amounts)
    %       'binMapForUse' is the lookup table that holds the bin index for each projection value
    %    --------------------------------
    
    % sampleResolution = 21;
    jointSample = [currFiltImg_A(hA+1:21:end-hA) currFiltImg_B(hB+1:21:end-hB)];
    fact = 100/numBins; halfFact = fact/2;
    shift = fact*rand(1) - halfFact; % = fact*(rand[0,1]-0.5) = fact*rand[-0.5,0.5]
    
    % edges is of length numBins+1
    percentiles = BoundBy(shift+(0:fact:100),0,100);
    percentiles(1) = 0; percentiles(end) = 100;
    edges = floor(prctile(double(jointSample),percentiles))-minFilterValue;
    edges(1) = 1; edges(end) = maxFilterRange; % these are the limits (might not appear in sample)
    binMapForUse = zeros(1,maxFilterRange,'uint32'); % c style ranging from 0 to numBins-1 (arbitrary initialization)
    for i = 1 : numBins
        binMapForUse(edges(i):edges(i+1)) = bitshift(i-1,nextBit-1);
    end
    
    binMapsForUse(filter_ind,:) = binMapForUse;
    
    % updates
    numBitsForBinning = log2(numBins); % = bits_for_filters(ind)
    nextBit = nextBit + numBitsForBinning;
    if (nextBit > 32)
        debugDisp(insideInfo,['current bit is: ' num2str(nextBit) ' - aborting loop']);
        break
    end    
end                       % end num_of_filters_for_coding on kernels loop

bitCountPerTable = nextBit - 1;
numKernelsPerAlternative = filter_ind;

