function [CFIs_A,CFIs_B] = ...
    GetResultsOfKernelApplication_VIDEO(A,B,TLboundary,width,classType,hA,wA,dA,hB,wB )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% the last 4 parameters are relevant to 'patch' mode, which means that A and B contain flattened
% patches, i.e. they are of the size: % size(A) = [width^2,hA*wA,dA]
% IMPORTANT: note that in this case, we will add at the bottom and right zero padding of (width-1)

patch_mode = 0;

if (~patch_mode)
    %% A] Padding
    % pre padding of '2*width' is for the fast and correct initialization of the first 'width' rows/cols
    preA = padarray(A,[TLboundary,TLboundary,TLboundary],0,'pre');
    preB = padarray(B,[TLboundary,TLboundary,TLboundary],0,'pre');
    
    % sizes
    [pad_hA,pad_wA,pad_fA,dA] = size(preA);
    [pad_hB,pad_wB,pad_fB,dB] = size(preB);
end


%% NEW CODE (Efrat)
if (width ~= 4)
    error('The only width supported is 4!')
end

if (~exist(fullfile(filepath_DATA,'Kernels.mat'),'file'))
    filters_ind=[...x   y   t   color % index 
                    0   0   0   1       5   % 1   DC
                    0   0   0   2       2   % 2
                    0   0   0   3       2   % 3
                    0   0   1   1       3   % 4   (0,0,1)
                    0   0   1   2       0   % 5
                    0   0   1   3       0   % 6
                    0   1   0   1       3   % 7   (0,1,0)
                    0   1   0   2       0   % 8
                    0   1   0   3       0   % 9
                    1   0   0   1       3   % 10  (1,0,0)
                    1   0   0   2       0   % 11
                    1   0   0   3       0   % 12
                    1   1   0   1       1   % 13  (1,1,0) only Gray
                    1   0   1   1       1   % 14  (1,0,1) only Gray
                    0   1   1   1       1   % 15  (0,1,1) only Gray
                    1   1   1   1       0   % 16  (1,1,1) only Gray
                    0   0   2   1       0   % 17  (0,0,2) only Gray
                    0   2   0   1       0   % 18  (0,2,0) only Gray
                    2   0   0   1       0   % 19  (2,0,0) only Gray
                    2   1   0   1       0   % 20-25   6 permutations of (1,2,0) only Gray
                    2   0   1   1       0   % ...
                    1   2   0   1       0
                    1   0   2   1       0
                    0   1   2   1       0
                    0   2   1   1       0
                    ];
                save Kernels.mat filters_ind;
else
    % Loading all the computed Kernels data with the numOfBits for each one
    load(fullfile(filepath_DATA,'Kernels.mat')); % loads filters_ind mat, which olds all the kernels indices and numOfBits
end
% A] Defining filters to use
filters_ind = filters_ind(:,1:4);

% B] Allocations
filter_num = size(filters_ind,1);
currFiltImgs_A = zeros(pad_hA,pad_wA,pad_fA,filter_num,classType);
currFiltImgs_B = zeros(pad_hB,pad_wB,pad_fB,filter_num,classType);
                
% C] Calculating the projections, i.e. "the real thing" :)
for filter_ind = 1:filter_num
    current_filter_ind = filters_ind(filter_ind,:);
    filter_mat = get_filter_from_ind_VIDEO(current_filter_ind(1:3));    % filter_mat is of size = (width,width,width,dA);
    current_color_ind = current_filter_ind(4);
    currFiltImgs_A(:,:,:,filter_ind) = GetKernelProjMEX_VIDEO(preA(:,:,:,current_color_ind), filter_mat);
    currFiltImgs_B(:,:,:,filter_ind) = GetKernelProjMEX_VIDEO(preB(:,:,:,current_color_ind), filter_mat);
end

%%
% Maxmimum allowed kernel projection value to be is 2^16-1.
% This is because when error is calculated, then kernel projections
% difference is powered by 2.
% Thus the error for maximum difference is ((2^16-1)^2) ~ 2^32 which is max
% int32. So to prevent overflow on next algorithm stages, I devide the
% results by 4
% if (width == 32)
%     MaxVal = 2^16-1;
%     MinVal = - MaxVal;
%     currFiltImgs_A = BoundBy((currFiltImgs_A / 4) , MinVal , MaxVal);
%     currFiltImgs_B = BoundBy((currFiltImgs_B / 4) , MinVal , MaxVal);
% end


%% E] Collection into cell arrays (which aren't contiguous in memory)
% CFIs_A = cell(1,length(filters_ind));
% CFIs_B = cell(1,length(filters_ind));
% for kernel_ind = 1 : length(filters_ind)
%     for frame_ind_A = TLboundary+1:pad_fA
%         CFIs_A{kernel_ind}{frame_ind_A-TLboundary} = currFiltImgs_A(TLboundary+1:end,TLboundary+1:end,frame_ind_A,kernel_ind);
%     end
%     for frame_ind_B = TLboundary+1:pad_fB
%         CFIs_B{kernel_ind}{frame_ind_B-TLboundary} = currFiltImgs_B(TLboundary+1:end,TLboundary+1:end,frame_ind_B,kernel_ind);
%     end
% end

CFIs_A = cell(1,pad_fA - TLboundary);
CFIs_B = cell(1,pad_fB - TLboundary);
for kernel_ind = 1 : length(filters_ind)
    for frame_ind_A = TLboundary+1:pad_fA
        CFIs_A{frame_ind_A-TLboundary}{kernel_ind} = currFiltImgs_A(TLboundary+1:end,TLboundary+1:end,frame_ind_A,kernel_ind);
    end
    for frame_ind_B = TLboundary+1:pad_fB
        CFIs_B{frame_ind_B-TLboundary}{kernel_ind} = currFiltImgs_B(TLboundary+1:end,TLboundary+1:end,frame_ind_B,kernel_ind);
    end
end

