function [AnnA2B,AnnB2A,...     % results, each is 3 times the size of A or B
    innerTimes,...              % profiling
    annErrorsA,annErrorsB]...   % of size A*(iterations+1) - hold errors after each iteration (and before beginning)
    = HashingSchemeNew_VIDEO(rgbA,rgbB,filepath_DATA,k,calcBnn,parameters,visualizationType,mask,patch_mode)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% % parameters.width                                -> Kernel width (for non-descriptor mode)
% % parameters.descriptor_params.descriptor_mode    -> Boolean saying is this patch or descriptor mode
% % parameters.descriptor_params.hA -> Height of image A descriptors
% % parameters.descriptor_params.wA -> Width of image A descriptors
% % parameters.descriptor_params.hB -> Height of image B descriptors
% % parameters.descriptor_params.wB -> Width of image B descriptors

% % parameters.dist
% % parameters.maxKernels -> Maximum number of kernels for calculations
%%  parameters.useTicsInside -> Do profiling
% % parameters.TLboundary -> Top left boundary
% % parameters.BRboundary -> Bottom right boundary
% % parameters.br_boundary_to_ignore -> Bottom right boudaries to ignore in the mapping calculation
% % parameters.numHashTables -> Number of hash tables to use
% % parameters.numHashs -> Width of hash table (should be 2)
% % parameters.insideInfo -> Print debug info


%% A] extracting input paramters

if (nargin < 6)
    error('HashingSchemeNew error: At least 5 input paraemters are needed');
end
width = parameters.width;
descriptor_mode = parameters.descriptor_params.descriptor_mode;
maxKernels = parameters.maxKernels;
useTics = parameters.useTicsInside;
TLboundary = parameters.TLboundary;
BRboundary = parameters.BRboundary;
br_boundary_to_ignore = parameters.br_boundary_to_ignore;
numTables = parameters.numHashTables;
numHashs = parameters.numHashs;
insideInfo = parameters.insideInfo;
CalcErrorImages = parameters.DebugCalcErrors;
dist = parameters.dist; % this is for CSH that keeps 'dist' pixels in x or y away from identity

rotation_invariant = 0;

if (~descriptor_mode)
    if (~patch_mode)
        % sizes
        [hA,wA,fA,dA] = size(rgbA);
        [hB,wB,fB,dB] = size(rgbB);
        A = zeros(hA,wA,fA,dA,'uint8');
        B = zeros(hB,wB,fB,dB,'uint8');
        
        if (dA == 3)
            for i = 1 : fA
                A(:,:,i,:) = rgb2ycbcr(reshape(rgbA(:,:,i,:), hA, wA, dA));
            end
        else
			error('not supported for video yet');
            A = rgbA; % Gray level image
        end
        
        if (dB == 3)
            for i = 1 : fB
                B(:,:,i,:) = rgb2ycbcr(reshape(rgbB(:,:,i,:), hB, wB, dB));
            end
        else
            error('not supported for video yet');
            B = rgbB; % Gray level image
        end
    else    % patch_mode

    end
end

if (dA ~= dB)
    error ('Image color channels must be the same: both RGB or both gray level images');
end

ColorChannels = dA;

visualizeCandidates = 0;
% visualizeHashTables = 0;

% Set wheather to create Walsh Hadamard projections in memory at the
% optimal way for algorithm run time(NumKernels*Img_Size matrix) or in the
% logical view (Number of images which are the projection results, as the number of WH kernels)

visualizeFilterDistributions = 0;
visualizeDenoising = false;

if (~exist('mask','var'))
    mask = [];
end
%% B] PARAMETERS/INITIALIZTIONS 1 -Internal (only for statistics)

% Collect statistics for analysis. Two kinds of information:
% 1) General statistics on number and quality of candidates: a histogram of the number of
%       candidates of each quality (=depth=numFilters)
% 2) Image (per pixel) statistics:
%       a) number of candidates image
%       b) quality of candidates image (average of depths / sum of depths)
%%% data structures for this:
bitCountPerTable = zeros(1,numTables);
%%%

% for debug - follow error improvements after each table
if (CalcErrorImages)
    annErrorsA = zeros(1,numTables+1); % since I also save the error before the first iteration
    if (calcBnn)
        annErrorsB = zeros(1,numTables+1); % since I also save the error before the first iteration
    else
        annErrorsB = [];
    end
end

%% C] PARAMETERS/INITIALIZTIONS 2 - Actual things

% depending on the patch width - how many kernels (maximum) do we want to compute

run_C_code = 1;

% prepare result matrices
d_mapping = 3; % Tree dimensions to the video: x, y ant t.
if (insideInfo)
    d_mapping = d_mapping + 1; % Add the error dimension in case of debug
end
AnnA2B = ones(hA,wA,fA,d_mapping,k,'int32');
if (calcBnn)
    AnnB2A = ones(hB,wB,fA,d_mapping,k,'int32');
else
    AnnB2A = [];
end
% Choose type
if (descriptor_mode == 0)
    if (width <= 8)
        classType = 'int16';
    else
        classType = 'int32';
    end
else
    classType = 'int16';%'int32';%
end
% explanations of the following variables: (symmetric for A and B)

% - nBestMapping32uA: of size like A, holds the current best found mapping which is by a FLAT index into B, that runs column after column
% - bestErrorsNewA  : of size like A, holds the approximated errors (GCK errors, not SSD errors) of the current best mapping

% alocate memory
nBestMapping32uA = zeros(hA,wA,fA,k,'uint32'); % current best mapping
bestErrorsNewA = zeros(hA,wA,fA,k,'uint32'); % current (GCK)-errors (approximation of SSD error)

% this calculates all filter images (returned in INT16 format)

% Output PCA projections is in class signle
if (~exist(fullfile(filepath_DATA,'applications.mat'),'file'))
    [currFiltImgs_A,currFiltImgs_B] = ...
        GetResultsOfKernelApplication_VIDEO(A,B,TLboundary,width,classType);
        save(fullfile(filepath_DATA,'applications.mat'), 'currFiltImgs_A', 'currFiltImgs_B');      
else
    load(fullfile(filepath_DATA,'applications.mat')); % loads currFiltImgs_A, currFiltImgs_B
end

debugDisp(insideInfo, '> done');

% initialize nBestMapping32uA with a random mapping (for each k)
for i = 1 : k
%     bestMappingNewA = (i+ max([hA,hB,fA]) + width + 1) * ones([hA,wA,fA],'uint32');
    bestMappingNewA = ( i + hA + hA*wA + width + 1) * ones([hA,wA,fA],'uint32');
    nBestMapping32uA(:,:,:,i) = uint32(bestMappingNewA)-1; % C syntax
    % initial errors
    bestErrorsNewA(:,:,:,i) = (2^32-1); % Max int, force candidate check
end
clear bestMappingNewA % (this was only the matlab version)


if (descriptor_mode)
    if (strcmp(classType,'int16'))
        currFiltImgs_A = int16(currFiltImgs_A);
        currFiltImgs_B = int16(currFiltImgs_B);
    else
        currFiltImgs_A = int32(currFiltImgs_A);
        currFiltImgs_B = int32(currFiltImgs_B);
    end
end

%% D] CALCULATING WH KERNELS - all done in advance (to be kept in memory)

% Trick - create big penalty for pixels at image margin (to avoid memory exceptions in propagation)
% Large val should be the maximum possible value of patch projection, which
% is the DC part: SUM(PATH_GRAY_LEVELS). - Since there are 256 gray levels
% (2^8), we get the following formula
if (descriptor_mode == 0)
    largeVal = 256*width*width*width-1; % This is max positive values range -1(e.g.: in 4 kernel bit, range is 4*4*4*256=16484, and largeval =  16383
else
    largeVal = MaxDescriptorIntVal;
end

% Prevent overflow in next algorithm stages cause projections are powered
% by 2 for error calculation (error is 32 bit unsigned int).
% So this is how we take care that error would not overflow
MaxVal = 2^16-1;
if (largeVal > MaxVal)
    largeVal = MaxVal;
end

% this is the main call (done in both calcBnn and ~calcBnn cases)
[currFiltImgs_A, currFiltImgs_B] = ModifyBoundaries_VIDEO(currFiltImgs_A, currFiltImgs_B,largeVal,0,width, mask);
    
%% F] Build walsh hadamard codes
% description: Build codes by using the WH projections calculated before
% and randmoally selecting a shift

debugDisp(insideInfo, '-----------  Build codes  --------------------------------------------------------------');

useWH_codes = 1; % Initial settings
if (descriptor_mode == 1)
    usePCA_codes = 1;
    useWH_codes = 0; % Do not use Walsh hadamard codes, rather use PCA codes
end
code_A = cell(1,numTables);
code_B = cell(1,numTables);

% CodesAddition = zeros(1,numTables);

for tableInd = 1 : numTables % (= iterations)      
    if (useWH_codes)
        % here I make a compromise (this is perfectly suitable only for A2B and not perfect for B2A)
        code_A_Img = cell(1,fA);
        code_B_Img = cell(1,fB);
        % preprocess - jointly (to the complete videos A and B) decide on filters and binMaps
        [bitCountPerTable(tableInd),binMapsForUse,codingFilters] = ...
            BuildWalshHadamardCodes_VIDEO1(filepath_DATA,tableInd,width,currFiltImgs_A, currFiltImgs_B, ...
            visualizeFilterDistributions,numTables,hA,wA,hB,wB,insideInfo,ColorChannels);
        
        % and use these per frame
        for i = 1 : fA     
            code_A_Img{i} = BuildWalshHadamardCodes_VIDEO2(tableInd,width,currFiltImgs_A{i},codingFilters,binMapsForUse,hA,wA,insideInfo,classType);
        end
        
        for i = 1 : fB     
            code_B_Img{i} = BuildWalshHadamardCodes_VIDEO2(tableInd,width,currFiltImgs_B{i},codingFilters,binMapsForUse,hB,wB,insideInfo,classType);
        end       
    else
        error('not supported for video');
    end
    code_A{tableInd} = code_A_Img;
    code_B{tableInd} = code_B_Img;  
end

clear tmp_code_A_Img code_A_Img tmp_code_B_Img code_B_Img

debugDisp(insideInfo, '--');

%% F] Memory convert
% description: Replace the filtering results images in memory with a more
% optimal arrangement in memory. This dramatically decreases candidate
% ranking run time.

debugDisp(insideInfo, '-----------  Memory convert --------------------------------------------------------------');

if (~descriptor_mode)
    % sizes
    numPixelsA = fA*wA*hA;
    numPixelsB = fB*wB*hB;
    CFI = currFiltImgs_A{1};
    Num_Filters = length(CFI);
    [Filter_H Filter_W] = size(CFI{1});
    
    CF_Vectors_A = zeros(Num_Filters,numPixelsA,classType);
    computedFilters = 1:Num_Filters;
    for i = 1 : fA
        % Allocate memory in the opposite way  to the one needed ([W*H , nKernels]),
        % THIS IS NOT A BUG, but in order to save time. doing so and making
        % transpose at the end is much faster than straight copy.
        tmp_CF_Vectors_A = zeros(Filter_H * Filter_W , Num_Filters, classType);
        for iteration = 1 : Num_Filters
            tmp_CF_Vectors_A(:,iteration) = currFiltImgs_A{i}{computedFilters(iteration)}(:);
        end
        
        % Transpose the memory block to get the desired memory organization
        CF_Vectors_A(:, (i-1)*wA*hA + 1 : i*wA*hA) = tmp_CF_Vectors_A';
    end
    
    if (visualizeCandidates == 0 && ~visualizeDenoising)
        clear currFiltImgs_A; % Clear previous representation of filter projections on image A
    end
    
    CF_Vectors_B = zeros(Num_Filters,numPixelsB,classType);
    
    for i = 1 : fB
        % Allocate memory in the opposite way  to the one needed ([W*H , nKernels]),
        % THIS IS NOT A BUG, but in order to save time. doing so and making
        % transpose at the end is much faster than straight copy.
        tmp_CF_Vectors_B = zeros(Filter_H * Filter_W , Num_Filters, classType);
        for iteration = 1 : Num_Filters
            tmp_CF_Vectors_B(:,iteration) = currFiltImgs_B{i}{computedFilters(iteration)}(:);
        end
        
        % Transpose the memory block to get the desired memory organization
        CF_Vectors_B(:, (i-1)*wB*hB + 1 : i*wB*hB) = tmp_CF_Vectors_B';
    end
    
    if (visualizeCandidates == 0 && ~visualizeDenoising)
        clear currFiltImgs_B; % Clear previous representation of filter projections on image B
    end
    
       
else %(~descriptor_mode)
    error('not implemented for video at the moment')
end

% Keeping nBestMapping32uA and bestErrorsNewA with 4 dimentions! 
% no data memory convert !!!
% if (k > 1)
%     error('not implemented for video at the moment')
%     nBestMapping32uA = reshape(nBestMapping32uA, hA*wA*fA,k)';
%     bestErrorsNewA = reshape(bestErrorsNewA, hA*wA*fA,k)';
% end

% if (useTics)
%     innerTimes.MemoryConvertTime = toc;
% end

debugDisp(insideInfo, '--');
debugDisp(insideInfo, '-----------  HUGE loop --------------------------------------------------------------');

%% G] HUGE (MAIN) loop
% description: We have at hand all the filters organized in an optimal way
% in the memory for the ranknig stage: matrix of [number_of_filter *  (Width* Height)]
% We construct one hash table at a time, from the previously aved code images.
%Then, we create candidates for each pixel (patch) and  choose the best candidate
% (comparing also the best candidate from the previous round.

maxBits = max(bitCountPerTable);

for tableInd = 1 : numTables % (= iterations)
    % 2] Hash into tables
    %    ----------------
%     if (useTics)
%         tic
%     end
    if (tableInd==1)
        hTables_A = [];hTables_B = [];
    end 
    debugDisp(insideInfo, '== Before HashCodesIntoTables');
    indices_A = zeros(hA,wA,fA,'uint32');
    for i = 1 : fA
        offset = (i-1)*hA*wA;
        [hTables_A,tmp_indices_A,~,~] = ...
            HashCodesIntoTables_VIDEO(hA,wA,fA,maxBits,bitCountPerTable,tableInd,numHashs,...
            code_A{tableInd}{i},hTables_A,insideInfo,offset);
        indices_A(:,:,i) = tmp_indices_A;
    end
    indices_B = zeros(hB,wB,fB,'uint32');
    for i = 1 : fB
        offset = (i-1)*hB*wB;
        [hTables_B,tmp_indices_B,~,~] = ...
            HashCodesIntoTables_VIDEO(hB,wB,fB,maxBits,bitCountPerTable,tableInd,numHashs,...
            code_B{tableInd}{i},hTables_B,insideInfo,offset);
        indices_B(:,:,i) = tmp_indices_B;
    end
    debugDisp(insideInfo, '== After HashCodesIntoTables');
    
    % 3] Candidate creation and ranking
    %    ------------------------------
    if (run_C_code)
        
        nHashsNoe16u = uint16(numHashs);
        
        % this indicates whether we are going upwards or downwards
        bDownwardFlag  = logical(mod(tableInd,2) == 1);
        % alg = @GCKCodebookPropagation16s_FilterVectors_VIDEO;
        alg = @GCKCodebookPropagation16s_FilterVectors_VIDEO_NEW;
        alg(bestErrorsNewA, nBestMapping32uA, CF_Vectors_A, CF_Vectors_B, indices_A, indices_B, hTables_A, hTables_B, ...
            nHashsNoe16u, bDownwardFlag, cast(dist,'uint32'));
        
        debugDisp(insideInfo, ['>>>>>>>>>                           done                       >>>>>>>' char(10)] );
    end % if (run_C_code)
    
end % End of Huge Loop

% debug
debugShow(insideInfo, 'bitCountPerTable',bitCountPerTable);


%% H] FINALIZE - arrange results in outputs
for i = 1 : k
    if (run_C_code)
        [AnnA2B(:,:,:,1,i),AnnA2B(:,:,:,2,i),AnnA2B(:,:,:,3,i)] = ind2sub([hB wB fB],nBestMapping32uA(:,:,:,i)+1);
    end
end

return
