% example usage for CSH_nn function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
close all
% clear all
clc

AddPaths

calcAll = 1;

fprintf('CSH algorithm example script!!!\r\n');
fprintf('*******************************\r\n');

%% settings
img1 = 'Saba1.bmp';
img2 = 'Saba2.bmp';

% First example: use internal defaults to show mapping
A = imread(img1);
B = imread(img2);
A = imresize(A,0.2); B = imresize(B,0.2); 
B = cat(3,flipud(A(:,:,1)),flipud(A(:,:,2)),flipud(A(:,:,3)));

% %
% load 'dressAB.mat'
% A = rgbImg2;
% B = rgbImg3;
% %
addpath('..\TreeCANN');
width = 8;
iterations = 4;
k = 1;
[hA wA dA] = size(A);
[hB wB dB] = size(B);
calcBNN = 0;
mask = [];

%% patch mode

A_grid = 1; B_grid = 1;
[A_patch_pos_X  A_patch_pos_Y] = extract_patch_position(size(A), width, A_grid);
[B_patch_pos_X  B_patch_pos_Y] = extract_patch_position(size(B), width, B_grid);

%% rotation invariance
grayA = double(rgb2gray(A));
grayB = double(rgb2gray(B));

% % for debug
% grayA = zeros(101);
% for i = 1 : 201
%     for j = 1 : 201
%         grayA(i,j) = 2*sqrt((i-101)^2+(j-101)^2);
%     end
% end
% figure; imshow(uint8(grayA))
% [hA,wA] = size(grayA);
% %

[Gx_A,Gy_A] = gradient(grayA);
[Gx_B,Gy_B] = gradient(grayB);

magsA = (Gx_A.^2+Gy_A.^2).^0.5;
magsB = (Gx_B.^2+Gy_B.^2).^0.5;


% the format: covers the range [0,2pi] starting at the north, going clockwise
orient_A = pi/2+atan(Gy_A./Gx_A)+pi*(Gx_A<=0); 
orient_B = pi/2+atan(Gy_B./Gx_B)+pi*(Gx_B<=0); 

% handle 'up' and 'down' derivatives explicitly (NaN problems)
orient_A(Gx_A==0) = pi*(Gy_A(Gx_A==0)>0);
orient_B(Gx_B==0) = pi*(Gy_B(Gx_B==0)>0);

% figure;  imagesc2(magsA,magsB,orient_A,orient_B);

nbins = width;

% create an image of pixel orientation bins
binsImg_A = 1 + floor(orient_A/(2*pi/nbins));
binsImg_B = 1 + floor(orient_B/(2*pi/nbins));  % figure; imagesc(binsImg_A)
        

% create a histograms of orientations image
if (calcAll || ~exist('hists.mat'))
    histImg_A = GetHistogramOrientations(binsImg_A,magsA,width,nbins);
    histImg_B = GetHistogramOrientations(binsImg_B,magsB,width,nbins);
    save 'hists.mat'
else
    load 'hists.mat'
end

% extracting the peaks (best bins) of the histograms
[~ , bestBins_A] = max(histImg_A,[],1);
bestBins_A = squeeze(bestBins_A);
[~ , bestBins_B] = max(histImg_B,[],1);
bestBins_B = squeeze(bestBins_B);

% take the weighted average of orientations in the winning bin (instead of just taking the bin center)
winnerOrientations_A = GetWinnerOrientationPerBlock(binsImg_A,bestBins_A,magsA,orient_A,width);
winnerOrientations_B = GetWinnerOrientationPerBlock(binsImg_B,bestBins_B,magsB,orient_B,width);

% visualization
imagesc2(winnerOrientations_A,binsImg_A,orient_A);
imagesc2(winnerOrientations_B,binsImg_B,orient_B);

% create array of invariant (rotated) patches
calcAll = 1;
if (calcAll || ~exist('including_A_patch.mat'))
    [A_patch,A_patch_RGB] = CreateArrayOfRotatedPatches(A,winnerOrientations_A,width);
    [B_patch,B_patch_RGB] = CreateArrayOfRotatedPatches(B,winnerOrientations_B,width);
    save ('including_A_patch.mat','A_patch','numPatches');
else
    load ('including_A_patch.mat','A_patch','numPatches');
end


keyboard
% A_patch = uint8(A_patch);

% while (1)
%     i = randi(hA-wM1);
%     j = randi(wA-wM1);
%     figure; imshow(A); hold on; plot(j,i,'*');
%     block = A(i:i+wM1,j:j+wM1,:);
%     ind = sub2ind([wA-wM1,hA-wM1],j,i);
%     rotBlock = A_patch(:,ind);
%     rblock = reshape(rotBlock,[width,width,3]);
%     imshow2(block,rblock);
% end
% save 'all.mat'
% keyboard

%%
if (~exist('A_patch'))
    separateChannels = int32(0); % choose 1 or 0 (1 will make CSH a little faster)
    
    [A_patch B_patch] = TreeCANN_extract_patches(uint8(rgb2ycbcr(A)), uint8(rgb2ycbcr(B)), width,...
        uint32(A_patch_pos_X) , uint32(A_patch_pos_Y) , uint32(B_patch_pos_X) , uint32(B_patch_pos_Y), separateChannels);
    [A_patch_RGB B_patch_RGB] = TreeCANN_extract_patches(uint8(A), uint8(B), width,...
        uint32(A_patch_pos_X) , uint32(A_patch_pos_Y) , uint32(B_patch_pos_X) , uint32(B_patch_pos_Y), separateChannels);
    % if separateChannels==1, then A_patch is of dimensions: [width^2,numPatches,3]
    %       else, it is: [3*width^2,numPatches], where each column contains all R, then all G, then all B
    %       in any case: the patches are taken ROW BY ROW
end

% we don't create patches for the width-1 rows and columns (hA*wA==size(A_patch,2))
hA = hA - width + 1;
wA = wA - width + 1;
hB = hB - width + 1;
wB = wB - width + 1;


fprintf('1. Runing CSH_nn PATCH MODE example: width = %d, iterations = %d, k = %d\r\n',width,iterations,k);
patch_mode = 1;
patch_params.hA = hA;
patch_params.wA = wA;
patch_params.dA = dA;
patch_params.hB = hB;
patch_params.wB = wB;
patch_params.dB = dB;
CSH_TIC = tic;
%%%%%% CSH RUN %%%%%%%
CSH_ann = CSH_nn(A_patch,B_patch,width,iterations,k,calcBNN,mask,patch_mode,patch_params); %CSH_ann is the size of the patch-image
%%%%%%%%%%%%%%%%%%%%%%
CSH_TOC = toc(CSH_TIC);
fprintf('   CSH_nn elapsed time: %.3f[s]\r\n' , CSH_TOC);
[errImg1, CSH_MeansRMS1] = PlotExampleResults(A,B,padarray(CSH_ann,[width-1,width-1],0,'post'),width,k);
set(gcf,'name',['patch mode - took ' num2str(CSH_TOC) ' seconds']);

% compute error image based on the patch representation
bMask = [];
interactive = 1;
[errImg2, CSH_MeansRMS2] = PlotExampleResults(A,B,CSH_ann,width,k,bMask,patch_mode,A_patch_RGB,B_patch_RGB,patch_params,interactive);
figure; imagesc(errImg2); title(['error mean: ' num2str(CSH_MeansRMS2)]);

%% regular mode

fprintf('2. Runing CSH_nn example: width = %d, iterations = %d, k = %d\r\n',width,iterations,k);
CSH_TIC = tic;
%%%%%% CSH RUN %%%%%%%
CSH_ann = CSH_nn(A,B,width,iterations,k,calcBNN,mask);
%%%%%%%%%%%%%%%%%%%%%%
CSH_TOC = toc(CSH_TIC);
fprintf('   CSH_nn elapsed time: %.3f[s]\r\n' , CSH_TOC);
PlotExampleResults(A,B,CSH_ann,width,k);
set(gcf,'name',['normal mode - took ' num2str(CSH_TOC) ' seconds']);

