% PREPARING ROTATED PATCHES
function [A_patch,A_patch_RGB,winnerOrientations_A,B_patch,B_patch_RGB,winnerOrientations_B] = ...
                                    PrepareRotatedPatchArrays(A,B,width,testName,calcAll)

grayA = double(rgb2gray(A));
grayB = double(rgb2gray(B));

% extracting gradients
[Gx_A,Gy_A] = gradient(grayA);
[Gx_B,Gy_B] = gradient(grayB);

magsA = (Gx_A.^2+Gy_A.^2).^0.5;
magsB = (Gx_B.^2+Gy_B.^2).^0.5;


% the format: covers the range [0,2pi] starting at the north, going clockwise
orient_A = pi/2+atan(Gy_A./Gx_A)+pi*(Gx_A<=0); 
orient_B = pi/2+atan(Gy_B./Gx_B)+pi*(Gx_B<=0); 

% handle 'up' and 'down' derivatives explicitly (NaN problems)
orient_A(Gx_A==0) = pi*(Gy_A(Gx_A==0)>0);
orient_B(Gx_B==0) = pi*(Gy_B(Gx_B==0)>0);

% DEBUG
% figure;  imagesc2(magsA,magsB,orient_A,orient_B);

nbins = width;

% create an image of pixel orientation bins
binsImg_A = 1 + floor(orient_A/(2*pi/nbins));
binsImg_B = 1 + floor(orient_B/(2*pi/nbins));  % figure; imagesc(binsImg_A)
        

% create A histograms of orientations image
filename = ['hists_' testName '.mat'];
if (calcAll || ~exist(filename))
    histImg_A = GetHistogramOrientations(binsImg_A,magsA,width,nbins);
    histImg_B = GetHistogramOrientations(binsImg_B,magsB,width,nbins);
    save (filename, 'histImg_A', 'histImg_B');
else
    load (filename, 'histImg_A', 'histImg_B');
end

% extracting the peaks (best bins) of the histograms
[~ , bestBins_A] = max(histImg_A,[],1);
bestBins_A = squeeze(bestBins_A);
[~ , bestBins_B] = max(histImg_B,[],1);
bestBins_B = squeeze(bestBins_B);

% take the weighted average of orientations in the winning bin (instead of just taking the bin center)
winnerOrientations_A = GetWinnerOrientationPerBlock(binsImg_A,bestBins_A,magsA,orient_A,width);
winnerOrientations_B = GetWinnerOrientationPerBlock(binsImg_B,bestBins_B,magsB,orient_B,width);

% visualization
imagesc2(winnerOrientations_A,binsImg_A,orient_A);
imagesc2(winnerOrientations_B,binsImg_B,orient_B);

% create array of invariant (rotated) patches
filename = ['including_A_patch_' testName '.mat'];
if (calcAll || ~exist(filename))
    [A_patch,A_patch_RGB] = CreateArrayOfRotatedPatches(A,winnerOrientations_A,width);
    [B_patch,B_patch_RGB] = CreateArrayOfRotatedPatches(B,winnerOrientations_B,width);
    save (filename,'A_patch','A_patch_RGB','B_patch','B_patch_RGB');
else
    load (filename,'A_patch','A_patch_RGB','B_patch','B_patch_RGB');
end
