function DCSH(params)
%% Preprocessing Parameters
loadPatchesPath         = params.loadPatchesPath;
ioPath                  = params.ioPath;
inputType               = params.inputType;
rgbImageName            = params.rgbImageName;
rgbdFileName            = params.rgbdFileName;
depthImageName          = params.depthImageName;
K                       = params.K;
gtRGBImageName          = params.gtRGBImageName;
logFolderName           = params.logFolderName;
noiseSigmaRGB           = params.noiseSigmaRGB;
nnRadius                = params.nnRadius;
bFitPlane               = params.bFitPlane;
patchResolution         = params.patchResolution;
xyWeight                = params.xyWeight;
depthWeight             = params.depthWeight;
socialMatch             = params.socialMatch;
loadCluster             = params.loadCluster;
loadClusterMatches      = params.loadClusterMatches;
knnCountImage           = params.knnCountImage;
showImages              = params.showImages;
showMatches             = params.showMatches;
showPCAOutput           = params.showPCAOutput;
showLowTextureMask      = params.showLowTextureMask;
numOfPlots              = params.numOfPlots;
clusterCount            = params.clusterCount;
knnCountClusters        = params.knnCountClusters;
cshParams.iterations    = params.cshIterations;
cshParams.calcBNN       = params.cshCalcBNN;
cshParams.mask          = params.cshMask;
cshParams.patchMode     = params.cshPatchMode;
cshParams.k             = knnCountClusters;
cshPatchParams.rotation_invariant  = params.cshRotationInvariant;
interpMethod            = params.interpMethod;
projectionRatioTH       = params.projectionRatioTH;
knnCountSampling        = params.knnCountSampling;
lowTextureThresh        = params.lowTextureThresh;
diffGaussianSigma       = params.diffGaussianSigma;
weightsGaussianSigma    = params.weightsGaussianSigma;
projectAllPatches       = params.projectAllPatches;
knnCountPCA = knnCountSampling;
%%
%Add \ at the suffix of the path if needed
ioPath          = FullPath(ioPath);
loadPatchesPath = FullPath(loadPatchesPath);

if exist(ioPath,'dir')~=7
    errordlg('The example data cluster could not be found');
    return;
end

if (~isdir(loadPatchesPath))
    mkdir(loadPatchesPath);
end

ijLogPath = [loadPatchesPath logFolderName '\'];
%Creating relevant directories for the specific image and parameters
if ~isdir([ijLogPath 'log\'])
    mkdir([ijLogPath 'log\']);
end

%% 1] Load/Calculate data of specific example
% e.g.: image, depthmap, K (intrinsics) as well as precomputed normals,
% homography arrays (ha) and not! Ha
if (exist([ijLogPath 'Preprocessing.mat'],'file'))
    load([ijLogPath 'Preprocessing.mat']);
else
    [projectionAreaArray,orientedHomographyArray,rgbImgAn,iRange,jRange,...
        normalMap,normalMapImage,depthZ] = mainDepthDirectNLM(ijLogPath,ioPath,K,rgbImageName,depthImageName,...
        rgbdFileName,bFitPlane,patchResolution,noiseSigmaRGB,nnRadius,inputType);
end

patchResolution2 = patchResolution^2;
diffGaussian0 = fspecial('gaussian',patchResolution + 1,diffGaussianSigma);
diffGaussian0 = reshape(diffGaussian0(1:patchResolution,1:patchResolution),[1 patchResolution2]);
diffGaussian0 = diffGaussian0/sum(diffGaussian0);
diffGaussian0 = reshape(diffGaussian0,[patchResolution patchResolution]);

%iRange - The x range of indexes that we are working on
%jRange - The y range of indexes that we are working on
jRangeSize = size(jRange,2);

if (exist('gtRGBImageName','var'))
    % loading original (noisy) RGB images
    rgbImgA = imread([ioPath gtRGBImageName]);
    
    if (size(rgbImgA,3) == 1)
        rgbImgA = repmat(rgbImgA,[1 1 3]);
    end
else
    rgbImgA = rgbImgAn;
end

%if it's a grey level image, duplicate it 3 times to be in same
%stracture of RGB image, but with values same as grey level image.
if (size(rgbImgAn,3) == 1)
    rgbImgAn = repmat(rgbImgAn,[1 1 3]);
end

rgbImgAn_double = double(rgbImgAn);

%% 2] Arrange loaded data
% For every pixel an homography is being saved,
% The homography is 3x3 matrix that is saved here as vector
% with 9 values
orientedHomographyArrayA = orientedHomographyArray;
focalLength = max(K(1,1),K(2,2));

projectionAreaArrayA      = projectionAreaArray;
projectionAreaArrayA_rect = projectionAreaArrayA(iRange,jRange);
clear projectionAreaArrayA;

orientedHomographyArrayA_rect = orientedHomographyArrayA(iRange,jRange,:);
clear orientedHomographyArrayA

% homographyMask_A is binary mask that keep only areas where homogrophies are known(required valid
% depth and consistent depth for normal calculation)
% the 3 dimention of orientedHomographyArrayA_rect contains the
% homography, check if the sum of values are valid (bigger then 0)
homographyMask_A = sum(abs(orientedHomographyArrayA_rect),3)>0;
patchCount = numel(homographyMask_A);

%%% column based:
% Change the orientedHomographyArrayA_rect to diffrent stractures
Ht = reshape(orientedHomographyArrayA_rect,[numel(homographyMask_A) 3 3]);
Ha_1 = reshape(Ht(:,:,1)',[3 * numel(homographyMask_A) 1]); %It's length is 3 times the number of pixels
Ha_2 = reshape(Ht(:,:,2)',[3 * numel(homographyMask_A) 1]); %It's length is 3 times the number of pixels
Ha_3 = reshape(Ht(:,:,3)',[3 * numel(homographyMask_A) 1]); %It's length is 3 times the number of pixels

Ha = zeros(patchCount * 3,3);
Ha(:,1) = Ha_1;
Ha(:,2) = Ha_2;
Ha(:,3) = Ha_3;

projectionWeightsBase0 = reshape(projectionAreaArrayA_rect,[numel(homographyMask_A) 1]);

% General : Here the CSH is running between an image and itself and not between
% two diffrent images
rgbImgAn_rect = rgbImgAn(iRange,jRange,:);
cshPatchParams.hA = size(rgbImgAn_rect,1); %height
cshPatchParams.wA = size(rgbImgAn_rect,2); %width
cshPatchParams.dA = size(rgbImgAn_rect,3); %depth
cshPatchParams.hB = cshPatchParams.hA;     %height
cshPatchParams.wB = cshPatchParams.wA;     %width
cshPatchParams.dB = cshPatchParams.dA;     %depth
orientedHomographyArrayA_rect_transpose = zeros(size(orientedHomographyArrayA_rect,2),size(orientedHomographyArrayA_rect,1),9);

for i=1:9
    orientedHomographyArrayA_rect_transpose(:,:,i) = orientedHomographyArrayA_rect(:,:,i)';
end

patchCountCSH = numel(homographyMask_A);
Ht_rows       = reshape(orientedHomographyArrayA_rect_transpose,[patchCountCSH 3 3]);
Ha_1rows      = reshape(Ht_rows(:,:,1)',[3 * patchCountCSH 1]);
Ha_2rows      = reshape(Ht_rows(:,:,2)',[3 * patchCountCSH 1]);
Ha_3rows      = reshape(Ht_rows(:,:,3)',[3 * patchCountCSH 1]);
Ha_rows       = zeros(patchCountCSH * 3,3);
Ha_rows(:,1)  = Ha_1rows;
Ha_rows(:,2)  = Ha_2rows;
Ha_rows(:,3)  = Ha_3rows;


%% 3] Arrange loaded data phase 2
% ij - index of the central pixel denoted as (xImg,yImg) of 1d array
% if not defined by the user, take the central pixel
% xImg - x value of the center point of the area we are interested to analyze
% yImg - y value of the center point of the area we are interested to analyze

if ~exist('xImg','var') && ~exist('yImg','var')
    yImg=floor((iRange(1) + iRange(end))/2);
    xImg=floor((jRange(1) + jRange(end))/2);
end

% boundaryX - Distance between the xImg to the left/right edge
% boundaryY - Distance between the yImg to the top/bottom edge
% if the boundries are not defined set Boundries values for the whole image

if ~exist('boundaryX','var') && ~exist('boundaryY','var')
    boundaryY = iRange(end) - yImg - 1;
    boundaryX = jRange(end) - xImg - 1;
end

ij = sub2ind(size(homographyMask_A(:,:,1)),yImg-iRange(1)+1,xImg-jRange(1)+1);

[ic, jc] = ind2sub(size(homographyMask_A),ij); %Getting the updated 2d indices
jcRange = jc+ (-boundaryX:(boundaryX-1));
icRange = ic+ (-boundaryY:(boundaryY-1));

% patchResolution - one dim of the patch size, the value 8 --> 8x8
% region: total area of patches (area of center pixels is 2 * b)
regionWidth  = 2 * boundaryX - 1 + patchResolution;
regionHeight = 2 * boundaryY - 1 + patchResolution;

%%% PCA structure init, the denoising is being done using PCA
outputHeight = regionHeight - 2 * (patchResolution - 1);
outputWidth  = regionWidth - 2 * (patchResolution - 1);
outputArea = outputHeight * outputWidth;

[Jgrid, Igrid] = meshgrid(0:(patchResolution - 1):((patchResolution - 1)^2),0:(patchResolution - 1):((patchResolution - 1)^2));

Jgrid = reshape(Jgrid,[1 patchResolution2]);
Igrid = reshape(Igrid,[1 patchResolution2]);
Jgrid = repmat(Jgrid,[outputArea 1]);
Igrid = repmat(Igrid,[outputArea 1]);

[J_offset, I_offset]  = meshgrid(patchResolution:patchResolution:(outputWidth * patchResolution), patchResolution:patchResolution:(outputHeight * patchResolution));
J_offset = repmat(reshape(J_offset,[outputArea 1]),[1 patchResolution2]);
I_offset = repmat(reshape(I_offset,[outputArea 1]),[1 patchResolution2]);

Jgrid = (Jgrid + J_offset)';   %%% patchResolution2 x outputArea
Igrid = (Igrid + I_offset)';   %%% patchResolution2 x outputArea

Jgrid = reshape(Jgrid,[patchResolution2 * outputArea 1]);
Igrid = reshape(Igrid,[patchResolution2 * outputArea 1]);
IJgrid = sub2ind([(patchResolution * 2 * boundaryY) (patchResolution * 2 * boundaryX)],Igrid,Jgrid);
height_patches_pca_array = patchResolution * 2 * boundaryY;
width_patches_pca_array  = patchResolution * 2 * boundaryX;

patchesPCA   = zeros(height_patches_pca_array, width_patches_pca_array,3,'uint8');
patchesPCA_R = zeros(height_patches_pca_array, width_patches_pca_array,'double');
patchesPCA_G = zeros(height_patches_pca_array, width_patches_pca_array,'double');
patchesPCA_B = zeros(height_patches_pca_array, width_patches_pca_array,'double');
weightsPCA   = zeros(height_patches_pca_array, width_patches_pca_array,'double');
outputPCA    = zeros(outputHeight,outputWidth,3,'double');

rangeAxc = (jRange(1)+jc-1)+(-boundaryX:(boundaryX-1));
rangeAyc = (iRange(1)+ic-1)+(-boundaryY:(boundaryY-1));

% in the following order: clean image, normals image, noisy image
patchAc = double(rgbImgA(rangeAyc,rangeAxc,:));
patchNormalMapImage = normalMapImage(rangeAyc,rangeAxc,:);
patchAc_noise = rgbImgAn_double(rangeAyc,rangeAxc,:);

patchAc_t = patchAc(patchResolution/2 : end - patchResolution/2,patchResolution/2 : end - patchResolution/2,:);
patchNormalMapImage_t = patchNormalMapImage(patchResolution/2 : end - patchResolution/2,patchResolution/2 : end - patchResolution/2,:);
patchAc_noise_t = patchAc_noise(patchResolution/2 : end - patchResolution/2,patchResolution/2 : end - patchResolution/2,:);

lowTextureMask = zeros(size(homographyMask_A));
dimMask = zeros(size(homographyMask_A));
textureMap = zeros(size(homographyMask_A));

[Ic,Jc] = meshgrid(icRange,jcRange); % Ic, Jc are matrices in the size of [IcRange x JcRnage]
% Ic hold the values of IcRange
% Jc hold the values of jCRange

Ic = reshape(Ic,[numel(Ic) 1]); %Make it 1 dim
Jc = reshape(Jc,[numel(Jc) 1]); %Make it 1 dim

outputRect = [(min(Ic) + patchResolution/2 - 1) (max(Ic) - patchResolution/2) (min(Jc) + patchResolution/2 - 1) (max(Jc) - patchResolution/2)];
maskDepthRegion = homographyMask_A(outputRect(1):outputRect(2),outputRect(3):outputRect(4));

if (numOfPlots > 0)
    intermediateOutputIndexes = floor(linspace(1,length(Ic),numOfPlots));
else
    intermediateOutputIndexes = 1:length(Ic);
end

%% 4] matching in image space (i.e. matching of standard square patches)
% General : All the patches are normalized in few view points (i.e 10)
% We also run another round of patch match in the image domain without using homographies and without
% using the normalized plains in order to find nearest neighbors, the reasons are:
% a) There are areas we don't have homographis so these patches (without any transformation) will be the
% only one avilable for matching
% b) When using homographis we sample the patch again in the image space,
% which result with interpolation, only the patches that didn't have any transformation poses the highest
% original quality.

% First: prepare patches for matching
preparedImagePatchesFilename = [ijLogPath 'preparedImagePatches.mat'];
disp('== matching in image space ==') %the standart square pathces
if ~exist(preparedImagePatchesFilename,'file')
    
    patchCount = numel(homographyMask_A);
    %%% row based:
    [Im, Jm] = meshgrid(1:size(homographyMask_A,1),1:size(homographyMask_A,2));
    Im = reshape(Im,[numel(Im) 1]);
    Jm = reshape(Jm,[numel(Jm) 1]);
    
    patchesImageCSH    = zeros(patchResolution2,patchCountCSH,3);
    knnImageIndexesCSH = zeros(size(homographyMask_A,1), size(homographyMask_A,2), knnCountImage,'uint32');
    disp('CSH: sampling patches')
    for ijm = 1:length(Im)
        i = Im(ijm);
        j = Jm(ijm);
        rangeAx = (jRange(1)+j-1)+(-(patchResolution/2):((patchResolution/2)-1));
        rangeAy = (iRange(1)+i-1)+(-(patchResolution/2):((patchResolution/2)-1));
        patchesImageCSH(:,ijm,:) = reshape(rgbImgAn_double(rangeAy,rangeAx,:),[patchResolution2 1 3]);
    end
    
    save(preparedImagePatchesFilename,'patchesImageCSH');
else
    load(preparedImagePatchesFilename,'patchesImageCSH');
end

% Now: do the matching using CSH
patchesImageCSHFilename = [ijLogPath 'patchesImageCSH.mat'];

disp('CSH: matching patches')
cshParams.k = knnCountImage; %k nearest neighbors
knnCount_t  = knnCountImage;
% Running the CSH with the same data for image A and for image B
% The input for CSH here are not images, they are array of patches.
CSH_ann = CSH_nn(patchesImageCSH,patchesImageCSH,patchResolution,cshParams.iterations,cshParams.k,cshParams.calcBNN,cshParams.mask,cshParams.patchMode,cshPatchParams);

%The stracture output of CSH_nn is : [ImageHeight x ImageWidth x number of indices x number of neigbours]
% Meaning that every pixel have k nearest neigbours that are
% identified with the indices of their center.

%%% row based:
knnImageIndexesCSH = uint32(squeeze((CSH_ann(:,:,2,:) - 1) * size(homographyMask_A,2) + CSH_ann(:,:,1,:)));
save(patchesImageCSHFilename,'knnImageIndexesCSH','knnCount_t');

%% 5] Cluster matching using homogramiphs using the diffrent view points
% First - calculate the representative viewpoints (clusters)
clusterFilename = [ijLogPath 'clusterData.mat'];
if loadCluster && exist(clusterFilename,'file')
    disp('Loading precomputed cluster')
    load(clusterFilename,'cluster');
else
    disp('computing cluster')
    cluster = CalculateClusterHomographies(normalMap,depthZ,K,patchResolution,clusterCount,...
        xyWeight,depthWeight);
    save(clusterFilename,'cluster');
end

% Second - perform CSH matching in each cluster
patchesClusterCSH = zeros(patchResolution2,patchCountCSH,3,cluster.count,'uint8');
knnClusterIndexesCSH = zeros(size(homographyMask_A,1), size(homographyMask_A,2), cluster.count * knnCountClusters,'uint32');

%K(1,3) represent the Camera X principal point
%K(2,3) represent the Camera Y principal point
%K(1:2,3) Ideally in the centre of the image
rangeAx =  round(K(1,3))+(-(patchResolution/2):((patchResolution/2)-1));
rangeAy =  round(K(2,3))+(-(patchResolution/2):((patchResolution/2)-1));
[pAx, pAy] = meshgrid(rangeAx,rangeAy);
pA = [reshape(pAx,[1 patchResolution2]) ; reshape(pAy,[1 patchResolution2]) ; ones(1,patchResolution2)];
ycbcrImgA = rgb2ycbcr(uint8(rgbImgAn));
knnCount_t = cshParams.k;
for h = 1:cluster.count
    patchesClusterCSHFilename = [ijLogPath 'patchesClusterCSH' num2str(h) '.mat'];
    % Optional loading for saving time, if the calculation was made
    % at previous run of the code
    if (loadClusterMatches && exist(patchesClusterCSHFilename,'file'))
        load(patchesClusterCSHFilename,'knnClusterIndexesCSH_t','knnClusterIndexesCSH');
    else
        H1 = cluster.H(:,:,h); %H1 Size[3x3xNumViewPoints], where 3x3 represent homography in the specific view point
        % and there are one homograpy for each view point
        %Building normalized patches - Na at the paper
        
        pN = H1 \ pA;  %Move the patch cordination to fronto-parallel space
        %The calculation is equivalent to inv(H1)*pA
        
        pN = pN./repmat(pN(3,:),[3 1]); %Force Homogeneous coordinates
        
        pB = Ha_rows * pN;
        pB = reshape(pB,[3 numel(pB) / 3]);
        pBx = pB(1,:) ./ pB(3,:);
        pBy = pB(2,:) ./ pB(3,:);
        pBx = reshape(pBx,[patchCountCSH numel(pBx) / patchCountCSH]);
        pBy = reshape(pBy,[patchCountCSH numel(pBy) / patchCountCSH]);
        pBx(isnan(pBx)) = 0;
        pBy(isnan(pBy)) = 0;
        
        % a. store normalized patches
        disp(['== Cluser ' num2str(h) ' out of ' num2str(cluster.count) ': sampling patches for CSH'])
        % Calculate the patches intensities for each homography using interpolation and for
        % each color channel (R,G,B)
        patchesClusterCSH(:,:,1,h) = uint8(vgg_interp2_omp(double(ycbcrImgA(:,:,1)),pBx',pBy',interpMethod)); %R
        patchesClusterCSH(:,:,2,h) = uint8(vgg_interp2_omp(double(ycbcrImgA(:,:,2)),pBx',pBy',interpMethod)); %G
        patchesClusterCSH(:,:,3,h) = uint8(vgg_interp2_omp(double(ycbcrImgA(:,:,3)),pBx',pBy',interpMethod)); %B
        
        disp('CSH: matching patches')
        % b. run CSH between each patches corresponds to the same
        % proto-typical plain (cluster)
        
        CSH_ann = CSH_nn(patchesClusterCSH(:,:,:,h),patchesClusterCSH(:,:,:,h),patchResolution,cshParams.iterations,cshParams.k,cshParams.calcBNN,cshParams.mask,cshParams.patchMode,cshPatchParams);
        
        clusterIndexes = (h - 1) * knnCount_t + (1:knnCount_t);
        
        %%% column based:
        knnClusterIndexesCSH_t = uint32(squeeze((CSH_ann(:,:,1,:) - 1) * size(homographyMask_A,1) + CSH_ann(:,:,2,:)));
        knnClusterIndexesCSH(:,:,clusterIndexes) = knnClusterIndexesCSH_t;
        save(patchesClusterCSHFilename,'knnClusterIndexesCSH_t','knnClusterIndexesCSH');
        
    end
end
disp('saving patchesClusterCSH.mat');
save([ijLogPath 'patchesClusterCSH.mat'],'knnClusterIndexesCSH');

%% taking cluster matches and resampling them to patchA space
HaAll = Ha;
normalizedPatchesFraction = zeros(size(homographyMask_A));

%% 6] Load the two types of computed matches: based on regular-image and based on clustered
disp('Loading CSH results on image');
load([ijLogPath 'patchesImageCSH.mat'],'knnImageIndexesCSH'); %Array of neigbors
% based on image space without homographis and without clustering
% (view points)

disp('Loading CSH results on cluster');
load([ijLogPath 'patchesClusterCSH.mat'],'knnClusterIndexesCSH');
numPixelsToReconstruct = length(Ic);

%% 7] Reconstruction Stage - a huge loop over pixels
for ijc = 1:numPixelsToReconstruct
    if mod(ijc,100)==0
        fprintf('Reconstruction loop: pixel %d out of %d (%.2f%%)\n',...
            ijc,numPixelsToReconstruct,100*ijc/numPixelsToReconstruct);
    end
    i = Ic(ijc);
    j = Jc(ijc);
    rangeAx = (jRange(1)+j-1)+(-(patchResolution/2):((patchResolution/2)-1));
    rangeAy = (iRange(1)+i-1)+(-(patchResolution/2):((patchResolution/2)-1));
    
    %% 7.1 Collecting arrays of NN patches
    if (homographyMask_A(i,j) == 0) %Define if there is valid homography at each pixel
        patchAn = rgbImgAn_double(rangeAy,rangeAx,:);
    else
        Hij = reshape(orientedHomographyArrayA_rect(i,j,:),[3 3]); % the Ha of the current i,j pixel
        [pAx,pAy] = meshgrid(rangeAx,rangeAy);
        pA = [reshape(pAx,[1 patchResolution2]) ; reshape(pAy,[1 patchResolution2]) ; ones(1,patchResolution2)];
        pN = Hij \ pA; %equivalent to pA * inv(Hij)
        pN = pN./repmat(pN(3,:),[3 1]); %Force homogeneousness cordinates
        
        %% grid for the gaussian mask sampling (this is for warping the Gaussian mask)
        patchAprojection = projectionAreaArrayA_rect(i,j);
        
        pG = focalLength * pN;
        gaussianRescaleFactor = sqrt(patchResolution2/patchAprojection);
        pGx = patchResolution/2 + reshape(pG(1,:)/gaussianRescaleFactor,[patchResolution patchResolution]);
        pGy = patchResolution/2 + reshape(pG(2,:)/gaussianRescaleFactor,[patchResolution patchResolution]);
        
        knnUniqueIndexes = unique(knnClusterIndexesCSH(i,j,:));
        if (socialMatch)
            numKnnUniqueIndexes = length(knnUniqueIndexes);
            socialIndexes = zeros(numKnnUniqueIndexes*(knnCountClusters * cshParams.k + 1),1);
            socialIndexes(1:numKnnUniqueIndexes) = knnUniqueIndexes;
            FirstInd = numKnnUniqueIndexes + 1;
            LastInd  = FirstInd + knnCountClusters * cshParams.k - 1;
            for k = 1:length(knnUniqueIndexes)
                [is,js] = ind2sub(size(homographyMask_A),knnUniqueIndexes(k));
                socialIndexes(FirstInd:LastInd) = squeeze(knnClusterIndexesCSH(is,js,:)); %Finding neigbours of the seed neuigbors
                FirstInd = LastInd  + 1;
                LastInd  = FirstInd + knnCountClusters * cshParams.k - 1;
            end
            knnUniqueIndexes = double(unique(socialIndexes)); %Remove all duplicates among the neigbors
        end
        
        knnIndexes = double(knnUniqueIndexes); %convert unit32 to double
        HaIndexes = (repmat((knnIndexes - 1) * 3,[1 3]) + repmat(1:3,[length(knnIndexes) 1]))';
        HaIndexes = reshape(HaIndexes,[numel(HaIndexes) 1]);
        Ha = HaAll(HaIndexes,:);
        patchCount = length(knnIndexes);
        pB = Ha * pN; % outputH = Ha*inv(Hij); (this includes all unfiltered NNs) (to be applied on pA representation)
        pB = reshape(pB,[3 numel(pB) / 3]);
        pBx = pB(1,:) ./ pB(3,:);
        pBy = pB(2,:) ./ pB(3,:);
        pBx = reshape(pBx,[patchCount numel(pBx) / patchCount]);
        pBy = reshape(pBy,[patchCount numel(pBy) / patchCount]);
        
        pBx(isnan(pBx)) = 0;
        pBy(isnan(pBy)) = 0;
        
        % ratios between current patch area and the areas of it NNs
        projectionWeightsBase = projectionWeightsBase0(knnIndexes);
        projectionRatio = (projectionWeightsBase / patchAprojection);
        
        patchAn = rgbImgAn_double(rangeAy,rangeAx,:);
        patchAn_vector = reshape(patchAn,[1 patchResolution2 3]);
        patchesAll_An = repmat(patchAn_vector,[patchCount 1 1]);
        
        %%%'sampling patches from noisy images'
        patchesAll_Bn = zeros(patchCount,patchResolution2,3);
        
        %the interp2 creates NAN values that are not cleaned, consider to replace them with zeros
        patchesAll_Bn(:,:,1) = vgg_interp2_omp(rgbImgAn_double(:,:,1),pBx,pBy,interpMethod); %R
        patchesAll_Bn(:,:,2) = vgg_interp2_omp(rgbImgAn_double(:,:,2),pBx,pBy,interpMethod); %G
        patchesAll_Bn(:,:,3) = vgg_interp2_omp(rgbImgAn_double(:,:,3),pBx,pBy,interpMethod); %R
        
        %% matches from P and sampling from P
        % in the normalized case - you need to warp the gaussian using the same homography that the patch undergoes
        diffGaussian = vgg_interp2_omp(double(diffGaussian0),pGx,pGy,'linear');
        diffGaussian(isnan(diffGaussian)) = 0;
        diffGaussian = reshape(diffGaussian,[1 patchResolution2]);
        diffGaussian = diffGaussian/sum(diffGaussian);
        diffGaussian = repmat(diffGaussian,[size(patchesAll_An,1) 1 3]);
        diff2 = sqrt(sum(sum(diffGaussian.*((patchesAll_An - patchesAll_Bn).^2) , 3), 2) / 3);
        diff2Indexes = 1:patchCount;
        diff2Mask = isnan(diff2) == 0;
        
        % don't take matches between world patches of extemely different size
        projectionMask = projectionRatio > projectionRatioTH;
        mask = logical(diff2Mask.*projectionMask);
        
        if (sum(mask) < (knnCountSampling * 0.5))
            mask = ones(size(mask));
        end
        
        diff2 = diff2(mask);
        diff2Indexes = diff2Indexes(mask)';
        projectionWeight = projectionWeightsBase(mask);
        % Sorting the patches generated from the clusters by error value
        weightsSorted = sortrows([diff2Indexes diff2 projectionWeight],2);
        
        %%% knnCount_t is knnCountFactor * knnCount
        knnCount_t = min(knnCountSampling,size(weightsSorted,1));
        weightsSorted = weightsSorted(1:knnCount_t,:);
        knnIndexes = weightsSorted(:,1);
        knnPatches = patchesAll_Bn(knnIndexes,:,:);
        
        if (showMatches)
            figure(2)
            for k = 1:min(knnCountSampling,50) % length(knnIndexes)
                subplot(5,10,k); imshow(uint8(reshape(knnPatches(k,:,:),[patchResolution patchResolution 3])));
                title(num2str(weightsSorted(k,2)));
            end
        end
        
    end % homography mask
    
    %% Denoising - using either PCA
    % perhaps cut out of this section the aranging of the final list of NN patches
    
    if (homographyMask_A(i,j))
        if (knnCount_t > knnCountPCA)
            knnCount_t = knnCountPCA;
            knnPatches = knnPatches(1:knnCount_t,:,:);
            weightsSorted = weightsSorted(1:knnCount_t,:);
            knnIndexes = knnIndexes(1:knnCount_t);
        end
    end
    
    clear knnPatchesMerged
    
    knnUniqueIndexes = knnImageIndexesCSH(i,j,:);
    if (socialMatch)
        socialIndexes = zeros(1+knnCountImage,1,knnCountImage);
        socialIndexes(1,1,:) = knnUniqueIndexes;
        for k = 1:knnCountImage %Number of values at knnUniqueIndexes
            % Change the 1D index back into 2D (js,is)
            [js,is] = ind2sub(jRangeSize,knnUniqueIndexes(k));
            socialIndexes(k+1,1,:) = knnImageIndexesCSH(is,js,:);
        end
        
        knnUniqueIndexes = unique(socialIndexes); %remove all duplicate neighbors from image plain neigbours
    end
    
    if (homographyMask_A(i,j))
        clear knnPatchesCSH
        knnPatchesCSH(:,:,1) = patchesImageCSH(:,knnUniqueIndexes,1)';
        knnPatchesCSH(:,:,2) = patchesImageCSH(:,knnUniqueIndexes,2)';
        knnPatchesCSH(:,:,3) = patchesImageCSH(:,knnUniqueIndexes,3)';
        
        patchAn_vector = reshape(patchAn,[1 patchResolution2 3]);
        patchesAll_An = repmat(patchAn_vector,[size(knnPatchesCSH,1) 1 1]);
        diffGaussian = fspecial('gaussian',patchResolution + 1,diffGaussianSigma);
        diffGaussian = reshape(diffGaussian(1:patchResolution,1:patchResolution),[1 patchResolution2]);
        diffGaussian = diffGaussian/sum(diffGaussian);
        diffGaussian = repmat(diffGaussian,[size(patchesAll_An,1) 1 3]);
        
        clear knnStatistics
        diff2 = sqrt(sum(sum(diffGaussian.*((patchesAll_An - knnPatchesCSH).^2) , 3), 2) / 3);
        diff2Indexes = 1:size(knnPatchesCSH,1);
        knnStatistics = [diff2Indexes' diff2];
        
        clear knnPatchesMerged
        knnPatchesMerged(:,:,1) = [knnPatches(:,:,1) ; knnPatchesCSH(:,:,1)];
        knnPatchesMerged(:,:,2) = [knnPatches(:,:,2) ; knnPatchesCSH(:,:,2)];
        knnPatchesMerged(:,:,3) = [knnPatches(:,:,3) ; knnPatchesCSH(:,:,3)];
        basicPatchesCount = size(knnPatchesCSH, 1);
    else
        knnPatchesMerged(:,:,1) = patchesImageCSH(:,knnUniqueIndexes,1)'; %R
        knnPatchesMerged(:,:,2) = patchesImageCSH(:,knnUniqueIndexes,2)'; %G
        knnPatchesMerged(:,:,3) = patchesImageCSH(:,knnUniqueIndexes,3)'; %B
    end
    
    if (homographyMask_A(i,j))
        diffMask = sum(isnan(knnPatchesMerged(:,:,1)),2) == 0;
        knnPatchesMerged = knnPatchesMerged(diffMask,:,:);
        mergedStatistics = [weightsSorted(:,2) ; knnStatistics(:,2)];
        mergedStatistics = mergedStatistics(diffMask,:);
        mergedStatistics = [(1:size(knnPatchesMerged,1))' mergedStatistics];
        
        knnCount_t = min(knnCount_t, size(knnPatchesMerged,1) - 1);
        mergedStatistics = sortrows(mergedStatistics,2);
        %%% ignoring duplicate first patch
        knnIndexes = mergedStatistics(1 + (1:knnCount_t),1);
        
        normalizedPatchesCount = size(knnPatches,1);
        normalizedPatchesFraction(i,j) = sum(knnIndexes <= normalizedPatchesCount) / numel(knnIndexes);
        knnPatches = knnPatchesMerged(knnIndexes,:,:);
    else
        diffMask = sum(isnan(knnPatchesMerged(:,:,1)),2) == 0;
        knnPatches = knnPatchesMerged(diffMask,:,:);
        knnCount_t = size(knnPatches,1);
        if (knnCount_t > knnCountPCA)
            knnCount_t = knnCountPCA;
            knnPatches = knnPatches(1:knnCount_t,:,:);
        end
    end
    
    %% PCA with diff weights
    if (projectAllPatches)
        weightsPCAProjection = exp(-weightsSorted(1:knnCount_t, 2) / noiseSigmaRGB);
    else
        % This is the case we use: clean and write to place only the current patch (not its NNs as well).
        % The assumption here is that the first (best) NN is the patch itself
        weightsPCAProjection = [1 ; zeros(knnCount_t - 1, 1)];
    end
    weightsPCAProjection = repmat(weightsPCAProjection / sum(weightsPCAProjection),[1 (patchResolution2 * 3)]);
    
    % stack KNN patches
    knnPatchesRGB = reshape(shiftdim(knnPatches,1),[patchResolution2 * 3 knnCount_t])';
    
    % weighted mean
    knnPatchesWeightedMean = mean(knnPatchesRGB);
    
    % get pca basis
    % first, make all the patches with mean zero
    knnPatchesRGBCentered = knnPatchesRGB - repmat(knnPatchesWeightedMean,[knnCount_t 1]);
    [PCA_basis,S,~] = svd(knnPatchesRGBCentered');
    %%% project all knn patches on reduced basis
    
    % find optimal dimension for reduction (based on L. Zhang, S. Vaddadi, H. Jin, and S. K. Nayar. Multiple view image denoising. In CVPR, 2009)
    singularValues = flipud(diag(S));
    sigmaNoiseSVD = 22;
    [~, indexSVD] = min(abs(cumsum(singularValues.^2) / (numel(knnPatchesRGB) * (sigmaNoiseSVD^2)) - 1));
    
    dim = length(singularValues) - indexSVD; % the number of dimension for PCA reduction
    dimMask(i,j) = dim;
    % dimension reduction and translating back to the original center of gravity
    projectionCoeffs = knnPatchesRGBCentered * PCA_basis(:, 1:dim);
    knnPatchesProjected = (projectionCoeffs * PCA_basis(:, 1:dim)') + repmat(knnPatchesWeightedMean, [knnCount_t 1]);
    patchPCA = sum(knnPatchesProjected .* weightsPCAProjection,1);
    patchPCA = reshape(patchPCA,[patchResolution,patchResolution,3]);
    
    numel_patchAn = numel(patchAn);
    flat_patch_dim = [1 numel_patchAn / 3]; %3 is the number of RGB channels
    stdPatchAn(1) = std(reshape(patchAn(:,:,1),flat_patch_dim)); %R
    stdPatchAn(2) = std(reshape(patchAn(:,:,2),flat_patch_dim)); %G
    stdPatchAn(3) = std(reshape(patchAn(:,:,3),flat_patch_dim)); %B
    
    stdPatchAn = mean(stdPatchAn);
    textureMap(i,j) = stdPatchAn;
    
    % in case of a 'low texture' patch use uniform (as opposed to 2D-Gaussian) spatial weighting
    if (stdPatchAn < lowTextureThresh)
        lowTextureMask(i,j) = 1;
        
        %%%	uniform reconstruction weights
        weightsGaussian = ones(patchResolution);
        weightsGaussian = weightsGaussian / numel(weightsGaussian);
    else
        
        %%% reconstruction weights
        weightsGaussian0 = fspecial('gaussian',patchResolution + 1,weightsGaussianSigma);
        weightsGaussian0 = weightsGaussian0(1:patchResolution,1:patchResolution);
        weightsGaussian0 = weightsGaussian0/sum(sum(weightsGaussian0));
        
        if (homographyMask_A(i,j))
            weightsGaussian = vgg_interp2_omp(double(weightsGaussian0),pGx,pGy,'linear');
            weightsGaussian(isnan(weightsGaussian)) = 0;
            weightsGaussian = weightsGaussian / sum(sum(weightsGaussian));
        else
            weightsGaussian = weightsGaussian0;
        end
    end
    
    [pca_j,pca_i] = ind2sub([2 * boundaryX, 2 * boundaryY],ijc);
    pca_i_indices = (pca_i - 1) * patchResolution + (1:patchResolution);
    pca_j_indices = (pca_j - 1) * patchResolution + (1:patchResolution);
    patchesPCA_R(pca_i_indices,pca_j_indices) = patchPCA(:,:,1);
    patchesPCA_G(pca_i_indices,pca_j_indices) = patchPCA(:,:,2);
    patchesPCA_B(pca_i_indices,pca_j_indices) = patchPCA(:,:,3);
    weightsPCA(pca_i_indices  , pca_j_indices,:) = weightsGaussian;
    
    % update PCA-OUTPUT figure visualization every ~2000 pixels
    if (sum(ijc == intermediateOutputIndexes) > 0)
        weightedPatches_R = patchesPCA_R .* weightsPCA;
        weightedPatches_G = patchesPCA_G .* weightsPCA;
        weightedPatches_B = patchesPCA_B .* weightsPCA;
        
        patchesPCA(:,:,1) = uint8(patchesPCA_R);
        patchesPCA(:,:,2) = uint8(patchesPCA_G);
        patchesPCA(:,:,3) = uint8(patchesPCA_B);
        
        weightsPCASum = reshape(sum(reshape(weightsPCA(IJgrid),[patchResolution2 outputArea]),1),[outputHeight outputWidth]);
        weightsPCAMask = weightsPCASum == 0;
        weightsPCASum(weightsPCAMask) = 1;  %%% to avoid dividing by zero
        
        outputPCA(:,:,1) = reshape(sum(reshape(weightedPatches_R(IJgrid),[patchResolution2 outputArea]),1),[outputHeight outputWidth]) ./ weightsPCASum;
        outputPCA(:,:,2) = reshape(sum(reshape(weightedPatches_G(IJgrid),[patchResolution2 outputArea]),1),[outputHeight outputWidth]) ./ weightsPCASum;
        outputPCA(:,:,3) = reshape(sum(reshape(weightedPatches_B(IJgrid),[patchResolution2 outputArea]),1),[outputHeight outputWidth]) ./ weightsPCASum;
        
        weightsPCASum(weightsPCAMask) = 0;
        maskPCA = weightsPCASum > 0;
        outputPCA(isnan(outputPCA)) = 0;
        
        if (showPCAOutput)
            figure(4)
            subplot(1,2,1); imshow(uint8(patchAn));
            title(['Noisy std ' num2str(stdPatchAn)])
            subplot(1,2,2); imshow(uint8(patchPCA));
            title('PCA')
            drawnow
        end
    end
    
    % update the detailed output visualization figure every ~2000 pixels
    if (sum(ijc == intermediateOutputIndexes) > 0)
        diff_results_vs_output = (patchAc_t - outputPCA).^2;
        rmsePCA_c   = sqrt(sum(sum(sum(repmat(maskPCA,[1 1 3]).*diff_results_vs_output)))/(sum(sum(maskPCA))*3));
        rmse_Denominator = (sum(sum(maskDepthRegion .* maskPCA))*3);
        rmse_Numerator   = repmat(maskDepthRegion .* maskPCA,[1 1 3]);
        if rmse_Denominator > 0 %To avoid NAN
            rmsePCA_depthRegion   = sqrt(sum(sum(sum(rmse_Numerator.*diff_results_vs_output)))/rmse_Denominator);
        else
            rmsePCA_depthRegion   = 0;
        end
        
        %%
        if (showImages)
            %Calculating RMSE (root mean square error)
            rmseImage_PCA  = sqrt(sum(diff_results_vs_output,3));
            figure(5)
            % Plotting results
            subplot(2,5,1); imshow(uint8(patchAc_t)); %Ground truth (clean) image
            title(['ref' num2str(ij)]);
            subplot(2,5,2); imshow(uint8(outputPCA)); %Output of noisy image in DCSH method
            title('new');
            subplot(2,5,3); imshow(uint8(patchAc_noise_t)); %Noisy input image
            title('noisy');
            subplot(2,5,4); imshow(rmseImage_PCA,[0 255]);
            title(['diff DCSH ' num2str(rmsePCA_c) ' ' num2str(rmsePCA_depthRegion)]);
            subplot(2,5,5); imshow(patchNormalMapImage_t);
            title('normal map');
            subplot(2,5,6); imshow(homographyMask_A(outputRect(1):outputRect(2),outputRect(3):outputRect(4)),[]);
            title('homography mask');
            subplot(2,5,7); imshow(textureMap(outputRect(1):outputRect(2),outputRect(3):outputRect(4)),[])
            title('texture map');
            subplot(2,5,8); imshow(dimMask(outputRect(1):outputRect(2),outputRect(3):outputRect(4)),[])
            title('PCA dimension map');
            subplot(2,5,9); imshow(normalizedPatchesFraction(outputRect(1):outputRect(2),outputRect(3):outputRect(4)),[0 1])
            title('Normalized Patch Fraction');
            colormap jet
            drawnow
            
            numNormalizedPatches = sum(sum(normalizedPatchesFraction > 0));
            
            if (showLowTextureMask)
                figure(6)
                subplot(121); imshow(lowTextureMask)
                subplot(122); imshow(dimMask,[])
            end
            
            if numNormalizedPatches > 0 %in case we didn't have normals so far, avoid division by zero
                fraction = sum(sum(normalizedPatchesFraction)) / numNormalizedPatches;
                title(['normalized fraction ' num2str(fraction)]);
            end
            
            colormap jet
            drawnow
        end %show images
    end
end

%% 8] Writing output results
imwrite(uint8(outputPCA),[ijLogPath 'log\image_dcsh.bmp'],'bmp');
imwrite(uint8(patchAc_noise_t),[ijLogPath 'log\image_noisy.bmp'],'bmp');
imwrite(uint8(rmseImage_PCA),colormap('jet'),[ijLogPath 'log\rmse_dcsh' num2str(rmsePCA_depthRegion) '.bmp'],'bmp');
imwrite(uint8(patchNormalMapImage_t),[ijLogPath 'log\normalMap.bmp'],'bmp');

homographyMaskBMP = uint8(homographyMask_A(outputRect(1):outputRect(2),outputRect(3):outputRect(4)) * 255);
textureMapBMP = textureMap(outputRect(1):outputRect(2),outputRect(3):outputRect(4));
textureMapBMP = uint8((double(textureMapBMP) / 60) * 255);

dimMaskBMP = dimMask(outputRect(1):outputRect(2),outputRect(3):outputRect(4));
dimMaskBMP = uint8((dimMaskBMP / 30) * 255);
normalizedFractionBMP = repmat(uint8(normalizedPatchesFraction(outputRect(1):outputRect(2),outputRect(3):outputRect(4)) * 255),[1 1]);

imwrite(homographyMaskBMP,[ijLogPath 'log\homographyMask.bmp'],'bmp');
imwrite(textureMapBMP,colormap('jet'),[ijLogPath 'log\textureMap.bmp'],'bmp');
imwrite(dimMaskBMP,colormap('jet'),[ijLogPath 'log\pcaDim.bmp'],'bmp');
imwrite(normalizedFractionBMP,colormap('jet'),[ijLogPath 'log\normalizedFraction' num2str(fraction) '.bmp'],'bmp');
close(gcf);

save([ijLogPath 'patch' num2str(ij) '_dcsh_' num2str(round(rmsePCA_c*100)) '.mat'],'ij','patchAc_t','outputPCA','patchAc_noise_t','rmsePCA_c','patchNormalMapImage_t');
disp('done');
end