function orientationArray = GetPatchArrayOrientations(normalizedPatchArrayImageRGB,homographyMask,patchResolution,iRange,jRange)

orientationArray = zeros(size(homographyMask,1),size(homographyMask,2),9);

%% PREPARING ROTATED PATCHES

grayA = double(rgb2gray(normalizedPatchArrayImageRGB));

% extracting gradients
[Gx_A,Gy_A] = gradient(grayA);

magsA = (Gx_A.^2+Gy_A.^2).^0.5;

% the format: covers the range [0,2pi] starting at the north, going clockwise
orient_A = pi/2+atan(Gy_A./Gx_A)+pi*(Gx_A<=0);

% handle 'up' and 'down' derivatives explicitly (NaN problems)
orient_A(Gx_A==0) = pi*(Gy_A(Gx_A==0)>0);


nbins = patchResolution;

% create an image of pixel orientation bins
binsImg_A = 1 + floor(orient_A/(2*pi/nbins));
%figure; imagesc(binsImg_A)


% create a histograms of orientations image
histImg_A = GetHistogramOrientations(binsImg_A,magsA,patchResolution,nbins);

% extracting the peaks (best bins) of the histograms
[~ , bestBins_A] = max(histImg_A,[],1);
bestBins_A = squeeze(bestBins_A);

% take the weighted average of orientations in the winning bin (instead of just taking the bin center)
winnerOrientations_A = GetWinnerOrientationPerBlock(binsImg_A,bestBins_A,magsA,orient_A,patchResolution);

orientationsSorted = sort(histImg_A,1,'descend');
orientation1 = squeeze(orientationsSorted(1,:,:));
orientation2 = squeeze(orientationsSorted(2,:,:));
dominantOrientationMap = (orientation1./ orientation2);

%%
cos_theta = cos(-winnerOrientations_A);
sin_theta = sin(-winnerOrientations_A);

cos_theta(~(dominantOrientationMap > 2)) = 1;
sin_theta(~(dominantOrientationMap > 2)) = 0;

orientationArray(iRange,jRange,1) = cos_theta;
orientationArray(iRange,jRange,2) = -sin_theta;
orientationArray(iRange,jRange,3) = 0;
orientationArray(iRange,jRange,4) = sin_theta;
orientationArray(iRange,jRange,5) = cos_theta;
orientationArray(iRange,jRange,6) = 0;
orientationArray(iRange,jRange,7) = 0;
orientationArray(iRange,jRange,8) = 0;
orientationArray(iRange,jRange,9) = 1;
end