function [homographyArray,projectionAreaArray,normalizedPatchArrayImageRGB_local] = ...
    NormalizePatchesLocal(K,X,Y,Z,normalMap,rgbImg,patchResolution,iRange,jRange)

%%% frontal patch at z=1 will occupy ~ 2*patchRadius*f
patchRadius = patchResolution/(2*max(K(1,1),K(2,2)));

n = normalMap;
[height,width,~] = size(n);

%%% allocate memory
homographyArray = zeros(height,width,9);
projectionAreaArray = zeros(height,width);
p2XAll = zeros(patchResolution*length(iRange),patchResolution*length(jRange),1,'double');
p2YAll = zeros(patchResolution*length(iRange),patchResolution*length(jRange),1,'double');

normalizedPatchArrayImageRGB_local = zeros(patchResolution*length(iRange),patchResolution*length(jRange),3,'uint8');
p2XAll_local = zeros(patchResolution*length(iRange),patchResolution*length(jRange),1,'double');
p2YAll_local = zeros(patchResolution*length(iRange),patchResolution*length(jRange),1,'double');

minI = iRange(1);
maxI = iRange(end);
minJ = jRange(1);
maxJ = jRange(end);

%%% normal rotation axis
theta = acos(n(:,:,3));

wx = -n(:,:,2);
wy = +n(:,:,1);
wz = zeros(height,width);

w = zeros(height,width,3);
w(:,:,1) = wx;
w(:,:,2) = wy;
w(:,:,3) = wz;
w_norm = sqrt(sum(w.^2,3));

wx = (wx./w_norm).*theta;
wy = (wy./w_norm).*theta;

wx(w_norm==0) = 0;
wy(w_norm==0) = 0;


%%% homography loop
height1 = patchResolution;
width1 = patchResolution;
patchGrid = [-patchRadius -patchRadius 0 ; patchRadius -patchRadius 0 ; patchRadius patchRadius 0 ;-patchRadius patchRadius 0]';

[p1X,p1Y] = meshgrid(linspace(-patchRadius,patchRadius,patchResolution),linspace(-patchRadius,patchRadius,patchResolution));
p1hGrid = zeros(3,width1*height1);
p1hGrid(1,:) = reshape(p1X,1,width1*height1);
p1hGrid(2,:) = reshape(p1Y,1,width1*height1);
p1hGrid(3,:) = ones(1,width1*height1);

img22 = rgbImg;
for i= minI:maxI
    if mod(i,10)==0
        fprintf('Normalized patches for line %d out of %d\n',i,maxI);
    end
    for j= minJ:maxJ

        if (sum(n(i,j,:))==0)
            %     'zero normal'
            continue
        end
        
        R = rodrigues([wx(i,j) wy(i,j) 0]');
        t = [X(i,j) ; Y(i,j) ; Z(i,j)];
        
        patchGridT = R*patchGrid + repmat(t,[1 4]);
        projectedPatch = K*patchGridT;
        projectedPatch = projectedPatch ./ repmat(projectedPatch(3,:),[3 1]);
        
        %%% checking ...
        rect(1) = min(projectedPatch(1,:));
        rect(2) = max(projectedPatch(1,:));
        rect(3) = min(projectedPatch(2,:));
        rect(4) = max(projectedPatch(2,:));
        
        rect = floor(rect);
        
        %%% out of boundaries
        if ((rect(1)<1)||(rect(3)<1)||(rect(2)>width)||(rect(4)>height))
            %'out'
            continue;
        else
            %%% nan
            if (sum(isnan(rect)))
                disp('nan');
                continue
            end
        end
        
        i0 = 1+(i-minI)*height1;
        j0 = 1+(j-minJ)*width1;
        
        %% ref depth normalization
        H = calcH(patchGrid,projectedPatch);
        homographyArray(i,j,:) = reshape(H,[1 9]);
        
        p2hGrid = H*p1hGrid;
        p2hGrid = p2hGrid./repmat(p2hGrid(3,:),3,1);
        p2XAll(i0:i0+height1-1,j0:j0+width1-1) = reshape(p2hGrid(1,:),height1,width1);
        p2YAll(i0:i0+height1-1,j0:j0+width1-1) = reshape(p2hGrid(2,:),height1,width1);
        
        %% local depth normalization
        patchCenterT = t;
        projectedPatchCenter = K*patchCenterT;
        projectedPatchCenter = projectedPatchCenter ./ repmat(projectedPatchCenter(3,:),[3 1]);
        projectedPatch_local = repmat(projectedPatchCenter,[1 4]) + Z(i,j) * (projectedPatch - repmat(projectedPatchCenter,[1 4]));

        H = calcH(patchGrid,projectedPatch_local);
        
        p2hGrid = H*p1hGrid;
        p2hGrid = p2hGrid./repmat(p2hGrid(3,:),3,1);
        p2XAll_local(i0:i0+height1-1,j0:j0+width1-1) = reshape(p2hGrid(1,:),height1,width1);
        p2YAll_local(i0:i0+height1-1,j0:j0+width1-1) = reshape(p2hGrid(2,:),height1,width1);
        
        %% calc projection area
        
        %%% projection diagonals
        Ax = projectedPatch(1,1);
        Ay = projectedPatch(2,1);
        Bx = projectedPatch(1,2);
        By = projectedPatch(2,2);
        Cx = projectedPatch(1,3);
        Cy = projectedPatch(2,3);
        Dx = projectedPatch(1,4);
        Dy = projectedPatch(2,4);
        
        %%% 0.5 abs(AC cross BD):
        projectionArea = 0.5 * abs((Cx - Ax) * (Dy - By) - (Dx - Bx) * (Cy - Ay));
        projectionAreaArray(i,j) = projectionArea;
        
    end
end

normalizedPatchArrayImageRGB_local(:,:,1) = uint8(interp2(double(img22(:,:,1)),p2XAll_local(:,:,1),p2YAll_local(:,:,1),'linear'));
normalizedPatchArrayImageRGB_local(:,:,2) = uint8(interp2(double(img22(:,:,2)),p2XAll_local(:,:,1),p2YAll_local(:,:,1),'linear'));
normalizedPatchArrayImageRGB_local(:,:,3) = uint8(interp2(double(img22(:,:,3)),p2XAll_local(:,:,1),p2YAll_local(:,:,1),'linear'));

end