function [normal, p0] = RansacPlaneFit(patch)
% f = #inliers / #points
% s = success probability
% n = number of points for each iteration
% t = inlier error threshhold (usually 1-3 pixels)

f = 0.7;
s = 0.99;
t = 0.005; %%% error thresh
n=6;

consensusSetBestCount = 0;
%%% iteration
iterationCount = max(1,ceil(log(1-s)/log(1-f^n)));

for i=1:iterationCount    

    indexes = ceil(rand(1,n)*length(patch));
    
    patchSamples = patch(indexes,:);
    
    [coeffPCA,scorePCA,rootsPCA] = princomp(patchSamples);
    
    basis = coeffPCA(:,1:2);
    normal = coeffPCA(:,3);
    
    p0 = mean(patchSamples,1);
    error = abs((patch-repmat(p0,[length(patch) 1]))*normal);
    %sse = sum(error.^2)
             
    consensusSetCount = sum(error<t);
      
    %%% if we got larger consensus Set
    if (consensusSetCount > max(consensusSetBestCount,6)) 
        
        consensusSetBestCount = consensusSetCount;
        
        pConsensus = patch(error<t,:);

        [coeffPCA,scorePCA,rootsPCA] = princomp(pConsensus);          
        normalBest = coeffPCA(:,3);                
        p0Best = mean(pConsensus,1);
    end 
end

%%% output
   normal = normalBest;     %%% plane normal
   p0 = p0Best;             %%% point on the plane
end