% Run DCSH Examples

% add paths using: 
if ~exist('CSH_nn','file')
    addpath(genpath(pwd));
end

%% Full Paramaters example : cup
% Basic parameters
params.loadPatchesPath    = 'c:\DCSH_Temp\'; % Path for Saving program temporal code calculations
params.ioPath             = '..\Testcases\cup\'; % ioPath   : Folder path of source files (requirements : RGB (Cleaned/Noisy),
                                                              % Depth file, Optional : groundTruthRGB)
params.inputType          = 'RGBD'; %'RGBD'\'TIFF'            % The type of file that hold the depth File
params.rgbImageName       = 'rgbImage.bmp';                   % Name and extension of the RGB Image
params.rgbdFileName       = 'RGBD.mat';                       % Name and extension of the 4-D RGBD file
params.depthImageName     = 'depthImage16bitFullRange.tiff';  % Name and extension Depth TIFF file
params.gtRGBImageName     = 'rgbImage.bmp';                   % Name of the ground RGB Image (optional)
params.logFolderName      =  'cup';                           % Name to be included in the log folder in order to allow future loading of pre computed parts
params.noiseSigmaRGB      = 25;                                % Value of the RGB channels noise
params.K                  = 100*[ 6.033482839192260,  0,   2.240000014305115;
                                  0,                  6.033482839192260,   2.999008283615112;
                                  0,                  0,   0.010000000000000];    %Camera intrinstic parameters
% Advanced parameters
params.patchResolution    = 8;%Patch size Size in pixels

% Normals
params.xyWeight = 0.5;      % Weights for clusters calcualation, giving lower weight to the x,y cordinates
params.depthWeight = 1;     % Weights for clusters calcualation of the depth
params.bFitPlane   = false; % In case the Image geomatry is flat, it's better to set it as true (for example paper)
                            % and then get bigger support of nnRadius
params.nnRadius    = 3;     % Normals calculation support

% combined RGB and depth matches
params.socialMatch = true;        % Run social CSH mechanism (Finding neigbours of seed nearest neigbours)
params.loadCluster = true;        % if exists load precomputed clusters homographies
params.loadClusterMatches = true; % if exists load precomputed clusters matches

% RGB only
params.knnCountImage  = 10;       % The number of seed nearest neighours in the image plain

% visualize
params.showImages    = true;      % Plotting : Ground truth (clean) image,Noisy picture after DCSH, Noisy Image, Root mean square error map, 
                     			  % Normal Map Homography mask, Texture map, PCA dimension map, Normalized Patch Fraction
params.showMatches   = false;     % show top 50 nearest patches for each pixel
params.showPCAOutput = false;     % Show original noisy patch vs. the patch after PCA output
params.showLowTextureMask = false; % Plotting map of area where patches standard deviation is less than lowTextureThresh  
params.numOfPlots    = 100;       % The number of times I would like to see visual output

% cluster
params.clusterCount     = 10;     % Number of prototypical view points
params.knnCountClusters = 10;     % Number of seed nearest neighours in each view point

% CSH
params.cshIterations = 6;         % CSH algorithm iterations
params.cshCalcBNN = 0;            % CSH calculated bnn (batch nearest neighbours) as well as ann (approximate nearest neighbours)  
params.cshMask = [];              % CSH mask on the target image, where to take the candidates from: 0 = use patch and 1 = hole
params.cshK = params.knnCountClusters; % CSH neighbors as in k-nearest-neighbors

% matching
params.interpMethod =  'nearest'; % 2-D interpolation parameters, alternatives : 'cubic','linear'
params.projectionRatioTH = 0.75;  % Threshold ratio between the a patch area and the areas of it NNs, NNs with ratio bigger than that will not be considered
params.knnCountSampling = 200;    % The final number of desired neighbours for each patch
params.sigmaN = 20;               % The sigma of the noise add to the clean image
params.lowTextureThresh = 30;     % Patch Standard deviation Thresh-hold
params.diffGaussianSigma = 4;     % Sigma of 2-D Gaussian that weights the patches difference that evaluate the nearest neighbour error

% reconstruction
params.weightsGaussianSigma = 2;  % Sigma of 2-D Gaussian that weights the patches for reconstruction
params.projectAllPatches = false; % if false use only one patch
defaults = Default_Values();
params = Complete_Parameters(params,defaults);
DCSH(params);

%% Examples with default values except the ones defined here
%% Book
defaults = Default_Values();
params.K = 100*[6.124821611090755, 0,                 3.039999926090240;
                0,                 6.124821611090755, 2.270617228746414;
                0,                 0,                 0.010000000000000];
params.ioPath  = '..\Testcases\book';
params.logFolderName      =  'book';
params = Complete_Parameters(params,defaults);
% main call
DCSH(params); % results are written at end of function

%% Dress
defaults = Default_Values();
params.K = 100 *[5.853483377687645, 0,                  3.039999926090240;
                 0,                 5.853483377687645,  2.270617228746414;
                 0                  0,                  0.010000000000000];
params.ioPath  = '..\Testcases\dress\';
params.logFolderName      =  'dress';
params = Complete_Parameters(params,defaults);
DCSH(params);

%% friends
defaults = Default_Values();
params.K = 100 *[5.917846256122830,     0,                  3.039999926090240;
                 0,                      5.917846256122830   2.270617228746414;
                 0,                      0,                  0.010000000000000];
params.ioPath  = '..\Testcases\friends\';
params.logFolderName      =  'friends';
params = Complete_Parameters(params,defaults);
DCSH(params);

%% mosaic
defaults = Default_Values();
params.K = 100 *   [6.033797933303760, 0,                   3.039999926090240;
                    0,                 6.033797933303760,   2.270617228746414;
                    0,                 0,                   0.010000000000000];
params.ioPath  = '..\Testcases\mosaic\';
params.logFolderName      =  'mosaic';
params.bFitPlane = true;
params = Complete_Parameters(params,defaults);
DCSH(params);

%% paper
defaults = Default_Values();
params.K =    1000*[2.037993487184235,     0,                   0.274000007629395;
                   0,                      2.037993487184235,   0.264839511871338;
                   0,                      0,                   0.001000000000000];
params.ioPath  = '..\Testcases\paper\';
params.logFolderName      =  'paper';
params.bFitPlane = true;
params = Complete_Parameters(params,defaults);
DCSH(params);

%% rock
defaults = Default_Values();
params.K =  100 * [6.656373004034042,      0,   2.560000181198120;
                   0,      6.656373004034042,   3.427438259124756;
                   0,                      0,   0.010000000000000];
params.ioPath  = '..\Testcases\rock\';
params.logFolderName      =  'rock';
params = Complete_Parameters(params,defaults);
DCSH(params);

%% tree
defaults = Default_Values();
params.K =  100 * [5.777565466318245,      0,                   2.240000014305115;
                   0,                      5.777565466318245,   2.999008283615112;
                   0,                      0,                   0.010000000000000];
params.ioPath  = '..\Testcases\rock\';
params.logFolderName      =  'tree';
params = Complete_Parameters(params,defaults);
DCSH(params);