% Function description : this code calculate the normal and homographis for
% an input image. 

function [projectionAreaArray,orientedHomographyArray,rgbImgAn,...
    iRange,jRange,normalMap,normalMapImage,depthZ] = mainDepthDirectNLM(ijLogPath,dataPath,K,RGBImageName,DepthImageName,...
    RGBDFileName,bFitPlane,patchResolution,noiseSigmaRGB,nnRadius,input_type)

%Inputs :
% testcaseName   - Name of test case, for example 'cup'/'book'
% noiseSigmaRGB  - Value of the RGB channels noise
% noiseGainDepth - Value of the Depth channel noise [0 0.25 0.5 1 2];
% RGBImageName   - Name and extension of the RGB Image (clean/noised)
% DepthImageName - Name and extension of the Depth Image
% bFitPlane      - In case the Image geomatry is flat, it's better to set it as true (for example paper)
% and then get bigger support of nnRadius

%%
depth16bitPath = [dataPath DepthImageName];
RGBPath        = [dataPath RGBImageName];
RGBDPath       = [dataPath RGBDFileName];

if (~isdir(dataPath))
    mkdir(dataPath)
end

switch input_type
    case 'RGBD'
        RGBD = load(RGBDPath);
        % Allows any variable name to hold the RGBD data
        RGBDFieldName=fieldnames(RGBD);
        RGBD   = RGBD.(RGBDFieldName{1});
        rgbImg = RGBD(:,:,1:3);
        depthZ = RGBD(:,:,4);
        [height,width,~] = size(rgbImg);
                
    case 'TIFF'
        %%% convert depth to meters
        rgbImg = imread(RGBPath,'bmp');
        depth16bit = double(imread(depth16bitPath));
        info   = imfinfo(depth16bitPath);
        spaces = find(isspace(info.ImageDescription));
        % User that chooses to store the depth map in TIFF file is extect
        % to write at this image description two values : minZ,maxZ of the
        % original depth map, in order to be able to retrive the original
        % values
        minZ = eval(info.ImageDescription(1:spaces(1)));
        maxZ = eval(info.ImageDescription(spaces(1)+1:end));
        depthZ     = Depth16bitToDepthZ(depth16bit,minZ,maxZ);
        [height,width,~] = size(rgbImg);
    otherwise
        warning('No relevant input type was found');
        return;
end

[X,Y,Z] = DepthZToXYZ(depthZ,K);
rgbImgAn  = AddNoiseRGB(rgbImg,noiseSigmaRGB);  % add noise to RGB image

%% Normals
if (bFitPlane)
    [planeX,planeY,planeZ,planeNormal] = FitDepthPlane(X,Y,Z,K);
    X = planeX;
    Y = planeY;
    Z = planeZ;
    depthZ = Z;
    
    normalMap = ones(height,width,3);
    normalMap(:,:,1) = planeNormal(1)*normalMap(:,:,1);
    normalMap(:,:,2) = planeNormal(2)*normalMap(:,:,2);
    normalMap(:,:,3) = planeNormal(3)*normalMap(:,:,3);
else
    % 'Fitting an Orthogonal Regression Using Principal Components Analysis'
    disp('calculating normals ...');
    bShowImage = false;
    normalMap = CalculateNormals(X,Y,Z,nnRadius,bShowImage);
end

normalMapImage = uint8(round(255*normalMap));

%% Normalize Patches
boundary = patchResolution;
iRange = boundary+1:height-boundary;
jRange = boundary+1:width-boundary;

disp('calculating normalized patches ...');
[homographyArray,projectionAreaArray,normalizedPatchArrayImageRGB_local] =  ...
    NormalizePatchesLocal(K,X,Y,Z,normalMap,rgbImgAn,patchResolution,iRange,jRange);

homographyMask = homographyArray(:,:,1)>0;

[orientationArray_local]  = GetPatchArrayOrientations(normalizedPatchArrayImageRGB_local,homographyMask,patchResolution,iRange,jRange);
orientedHomographyArray   = homographyArray;

disp('calculating homographies ...');
for ihr = 	iRange
    if mod(ihr,10)==0
        fprintf('homographies for line %d out of %d\n',ihr,iRange(end));
    end
    for jhr = jRange
        H = reshape(homographyArray(ihr,jhr,:),[3 3]);
        R = reshape(orientationArray_local(ihr,jhr,:),[3 3]);
        HR = H*R; %R - in plain rotation matrix, that force the gradient to be up
        orientedHomographyArray(ihr,jhr,:) = reshape(HR,[1 9]);
    end
end

orientedHomographyArray = RenormalizePatchesWithOrientationLocal(homographyArray,orientationArray_local,iRange,jRange);

save([ijLogPath 'Preprocessing.mat'],'projectionAreaArray',...
    'orientedHomographyArray','homographyArray','rgbImgAn','patchResolution','iRange','jRange','K','normalMap','normalMapImage','depthZ');
end